# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/gcs.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1asrc/ray/protobuf/gcs.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\"m\n\x08GcsEntry\x12\x37\n\x0b\x63hange_mode\x18\x01 \x01(\x0e\x32\x16.ray.rpc.GcsChangeModeR\nchangeMode\x12\x0e\n\x02id\x18\x02 \x01(\x0cR\x02id\x12\x18\n\x07\x65ntries\x18\x03 \x03(\x0cR\x07\x65ntries\"+\n\x0fObjectTableData\x12\x18\n\x07manager\x18\x01 \x01(\x0cR\x07manager\"\xea\t\n\x0e\x41\x63torTableData\x12\x19\n\x08\x61\x63tor_id\x18\x01 \x01(\x0cR\x07\x61\x63torId\x12\x1b\n\tparent_id\x18\x02 \x01(\x0cR\x08parentId\x12\x42\n\x1e\x61\x63tor_creation_dummy_object_id\x18\x03 \x01(\x0cR\x1a\x61\x63torCreationDummyObjectId\x12\x15\n\x06job_id\x18\x04 \x01(\x0cR\x05jobId\x12\x38\n\x05state\x18\x06 \x01(\x0e\x32\".ray.rpc.ActorTableData.ActorStateR\x05state\x12!\n\x0cmax_restarts\x18\x07 \x01(\x03R\x0bmaxRestarts\x12!\n\x0cnum_restarts\x18\x08 \x01(\x04R\x0bnumRestarts\x12*\n\x07\x61\x64\x64ress\x18\t \x01(\x0b\x32\x10.ray.rpc.AddressR\x07\x61\x64\x64ress\x12\x35\n\rowner_address\x18\n \x01(\x0b\x32\x10.ray.rpc.AddressR\x0cownerAddress\x12\x1f\n\x0bis_detached\x18\x0b \x01(\x08R\nisDetached\x12\x12\n\x04name\x18\x0c \x01(\tR\x04name\x12\x1c\n\ttimestamp\x18\r \x01(\x01R\ttimestamp\x12\x44\n\x10resource_mapping\x18\x0f \x03(\x0b\x32\x19.ray.rpc.ResourceMapEntryR\x0fresourceMapping\x12\x10\n\x03pid\x18\x10 \x01(\rR\x03pid\x12L\n\x13\x66unction_descriptor\x18\x11 \x01(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x12\x66unctionDescriptor\x12#\n\rray_namespace\x18\x13 \x01(\tR\x0crayNamespace\x12\x1d\n\nstart_time\x18\x14 \x01(\x04R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x15 \x01(\x04R\x07\x65ndTime\x12\x34\n\x16serialized_runtime_env\x18\x16 \x01(\tR\x14serializedRuntimeEnv\x12\x1d\n\nclass_name\x18\x17 \x01(\tR\tclassName\x12\x39\n\x0b\x64\x65\x61th_cause\x18\x18 \x01(\x0b\x32\x18.ray.rpc.ActorDeathCauseR\ndeathCause\x12]\n\x12required_resources\x18\x1c \x03(\x0b\x32..ray.rpc.ActorTableData.RequiredResourcesEntryR\x11requiredResources\x12\x1c\n\x07node_id\x18\x1d \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x31\n\x12placement_group_id\x18\x1e \x01(\x0cH\x01R\x10placementGroupId\x88\x01\x01\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"a\n\nActorState\x12\x18\n\x14\x44\x45PENDENCIES_UNREADY\x10\x00\x12\x14\n\x10PENDING_CREATION\x10\x01\x12\t\n\x05\x41LIVE\x10\x02\x12\x0e\n\nRESTARTING\x10\x03\x12\x08\n\x04\x44\x45\x41\x44\x10\x04\x42\n\n\x08_node_idB\x15\n\x13_placement_group_id\"~\n\x0e\x45rrorTableData\x12\x15\n\x06job_id\x18\x01 \x01(\x0cR\x05jobId\x12\x12\n\x04type\x18\x02 \x01(\tR\x04type\x12#\n\rerror_message\x18\x03 \x01(\tR\x0c\x65rrorMessage\x12\x1c\n\ttimestamp\x18\x04 \x01(\x01R\ttimestamp\"\x9d\x01\n\x0cScheduleData\x12L\n\rschedule_plan\x18\x01 \x03(\x0b\x32\'.ray.rpc.ScheduleData.SchedulePlanEntryR\x0cschedulePlan\x1a?\n\x11SchedulePlanEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01\"\x9f\x01\n\x11ProfileEventEntry\x12\x1d\n\nstart_time\x18\x01 \x01(\x03R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x02 \x01(\x03R\x07\x65ndTime\x12\"\n\nextra_data\x18\x03 \x01(\tH\x00R\textraData\x88\x01\x01\x12\x1d\n\nevent_name\x18\x04 \x01(\tR\teventNameB\r\n\x0b_extra_data\"\xb5\x01\n\rProfileEvents\x12%\n\x0e\x63omponent_type\x18\x01 \x01(\tR\rcomponentType\x12!\n\x0c\x63omponent_id\x18\x02 \x01(\x0cR\x0b\x63omponentId\x12&\n\x0fnode_ip_address\x18\x03 \x01(\tR\rnodeIpAddress\x12\x32\n\x06\x65vents\x18\x04 \x03(\x0b\x32\x1a.ray.rpc.ProfileEventEntryR\x06\x65vents\"\x8c\x04\n\x0fTaskStateUpdate\x12\x1c\n\x07node_id\x18\x01 \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x36\n\x15pending_args_avail_ts\x18\x02 \x01(\x03H\x01R\x12pendingArgsAvailTs\x88\x01\x01\x12@\n\x1apending_node_assignment_ts\x18\x03 \x01(\x03H\x02R\x17pendingNodeAssignmentTs\x88\x01\x01\x12\x38\n\x16submitted_to_worker_ts\x18\x04 \x01(\x03H\x03R\x13submittedToWorkerTs\x88\x01\x01\x12\"\n\nrunning_ts\x18\x05 \x01(\x03H\x04R\trunningTs\x88\x01\x01\x12$\n\x0b\x66inished_ts\x18\x06 \x01(\x03H\x05R\nfinishedTs\x88\x01\x01\x12 \n\tfailed_ts\x18\x07 \x01(\x03H\x06R\x08\x66\x61iledTs\x88\x01\x01\x12 \n\tworker_id\x18\x08 \x01(\x0cH\x07R\x08workerId\x88\x01\x01\x42\n\n\x08_node_idB\x18\n\x16_pending_args_avail_tsB\x1d\n\x1b_pending_node_assignment_tsB\x19\n\x17_submitted_to_worker_tsB\r\n\x0b_running_tsB\x0e\n\x0c_finished_tsB\x0c\n\n_failed_tsB\x0c\n\n_worker_id\"\xd8\x02\n\nTaskEvents\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\x12%\n\x0e\x61ttempt_number\x18\x02 \x01(\x05R\rattemptNumber\x12\x38\n\ttask_info\x18\x03 \x01(\x0b\x32\x16.ray.rpc.TaskInfoEntryH\x00R\x08taskInfo\x88\x01\x01\x12\x42\n\rstate_updates\x18\x04 \x01(\x0b\x32\x18.ray.rpc.TaskStateUpdateH\x01R\x0cstateUpdates\x88\x01\x01\x12\x42\n\x0eprofile_events\x18\x05 \x01(\x0b\x32\x16.ray.rpc.ProfileEventsH\x02R\rprofileEvents\x88\x01\x01\x12\x15\n\x06job_id\x18\x06 \x01(\x0cR\x05jobIdB\x0c\n\n_task_infoB\x10\n\x0e_state_updatesB\x11\n\x0f_profile_events\"\xd4\x01\n\rTaskEventData\x12\x39\n\x0e\x65vents_by_task\x18\x01 \x03(\x0b\x32\x13.ray.rpc.TaskEventsR\x0c\x65ventsByTask\x12\x44\n\x1fnum_profile_task_events_dropped\x18\x03 \x01(\x05R\x1bnumProfileTaskEventsDropped\x12\x42\n\x1enum_status_task_events_dropped\x18\x04 \x01(\x05R\x1anumStatusTaskEventsDropped\"@\n\x11ResourceTableData\x12+\n\x11resource_capacity\x18\x01 \x01(\x01R\x10resourceCapacity\"\xda\x01\n\x12\x41vailableResources\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x64\n\x13resources_available\x18\x02 \x03(\x0b\x32\x33.ray.rpc.AvailableResources.ResourcesAvailableEntryR\x12resourcesAvailable\x1a\x45\n\x17ResourcesAvailableEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xd4\x05\n\x0bGcsNodeInfo\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x30\n\x14node_manager_address\x18\x02 \x01(\tR\x12nodeManagerAddress\x12,\n\x12raylet_socket_name\x18\x03 \x01(\tR\x10rayletSocketName\x12\x37\n\x18object_store_socket_name\x18\x04 \x01(\tR\x15objectStoreSocketName\x12*\n\x11node_manager_port\x18\x05 \x01(\x05R\x0fnodeManagerPort\x12.\n\x13object_manager_port\x18\x06 \x01(\x05R\x11objectManagerPort\x12\x37\n\x05state\x18\x07 \x01(\x0e\x32!.ray.rpc.GcsNodeInfo.GcsNodeStateR\x05state\x12\x32\n\x15node_manager_hostname\x18\x08 \x01(\tR\x13nodeManagerHostname\x12.\n\x13metrics_export_port\x18\t \x01(\x05R\x11metricsExportPort\x12Q\n\x0fresources_total\x18\x0b \x03(\x0b\x32(.ray.rpc.GcsNodeInfo.ResourcesTotalEntryR\x0eresourcesTotal\x12\x1b\n\tnode_name\x18\x0c \x01(\tR\x08nodeName\x12\"\n\rstart_time_ms\x18\x17 \x01(\x04R\x0bstartTimeMs\x12\x1e\n\x0b\x65nd_time_ms\x18\x18 \x01(\x04R\tendTimeMs\x1a\x41\n\x13ResourcesTotalEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"#\n\x0cGcsNodeState\x12\t\n\x05\x41LIVE\x10\x00\x12\x08\n\x04\x44\x45\x41\x44\x10\x01\"\xa4\x07\n\x0bJobsAPIInfo\x12\x16\n\x06status\x18\x01 \x01(\tR\x06status\x12\x1e\n\nentrypoint\x18\x02 \x01(\tR\nentrypoint\x12\x1d\n\x07message\x18\x03 \x01(\tH\x00R\x07message\x88\x01\x01\x12\"\n\nerror_type\x18\x04 \x01(\tH\x01R\terrorType\x88\x01\x01\x12\"\n\nstart_time\x18\x05 \x01(\x04H\x02R\tstartTime\x88\x01\x01\x12\x1e\n\x08\x65nd_time\x18\x06 \x01(\x04H\x03R\x07\x65ndTime\x88\x01\x01\x12>\n\x08metadata\x18\x07 \x03(\x0b\x32\".ray.rpc.JobsAPIInfo.MetadataEntryR\x08metadata\x12-\n\x10runtime_env_json\x18\x08 \x01(\tH\x04R\x0eruntimeEnvJson\x88\x01\x01\x12\x33\n\x13\x65ntrypoint_num_cpus\x18\t \x01(\x01H\x05R\x11\x65ntrypointNumCpus\x88\x01\x01\x12\x33\n\x13\x65ntrypoint_num_gpus\x18\n \x01(\x01H\x06R\x11\x65ntrypointNumGpus\x88\x01\x01\x12`\n\x14\x65ntrypoint_resources\x18\x0b \x03(\x0b\x32-.ray.rpc.JobsAPIInfo.EntrypointResourcesEntryR\x13\x65ntrypointResources\x12>\n\x19\x64river_agent_http_address\x18\x0c \x01(\tH\x07R\x16\x64riverAgentHttpAddress\x88\x01\x01\x12)\n\x0e\x64river_node_id\x18\r \x01(\tH\x08R\x0c\x64riverNodeId\x88\x01\x01\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x46\n\x18\x45ntrypointResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\n\n\x08_messageB\r\n\x0b_error_typeB\r\n\x0b_start_timeB\x0b\n\t_end_timeB\x13\n\x11_runtime_env_jsonB\x16\n\x14_entrypoint_num_cpusB\x16\n\x14_entrypoint_num_gpusB\x1c\n\x1a_driver_agent_http_addressB\x11\n\x0f_driver_node_id\"\xf0\x02\n\x0cJobTableData\x12\x15\n\x06job_id\x18\x01 \x01(\x0cR\x05jobId\x12\x17\n\x07is_dead\x18\x02 \x01(\x08R\x06isDead\x12\x1c\n\ttimestamp\x18\x03 \x01(\x03R\ttimestamp\x12*\n\x11\x64river_ip_address\x18\x04 \x01(\tR\x0f\x64riverIpAddress\x12\x1d\n\ndriver_pid\x18\x05 \x01(\x03R\tdriverPid\x12*\n\x06\x63onfig\x18\x06 \x01(\x0b\x32\x12.ray.rpc.JobConfigR\x06\x63onfig\x12\x1d\n\nstart_time\x18\x07 \x01(\x04R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x08 \x01(\x04R\x07\x65ndTime\x12\x1e\n\nentrypoint\x18\t \x01(\tR\nentrypoint\x12\x34\n\x08job_info\x18\n \x01(\x0b\x32\x14.ray.rpc.JobsAPIInfoH\x00R\x07jobInfo\x88\x01\x01\x42\x0b\n\t_job_info\"\xe7\x04\n\x0fWorkerTableData\x12\x19\n\x08is_alive\x18\x01 \x01(\x08R\x07isAlive\x12\x37\n\x0eworker_address\x18\x02 \x01(\x0b\x32\x10.ray.rpc.AddressR\rworkerAddress\x12\x1c\n\ttimestamp\x18\x03 \x01(\x03R\ttimestamp\x12\x34\n\x0bworker_type\x18\x05 \x01(\x0e\x32\x13.ray.rpc.WorkerTypeR\nworkerType\x12I\n\x0bworker_info\x18\x06 \x03(\x0b\x32(.ray.rpc.WorkerTableData.WorkerInfoEntryR\nworkerInfo\x12M\n\x17\x63reation_task_exception\x18\x12 \x01(\x0b\x32\x15.ray.rpc.RayExceptionR\x15\x63reationTaskException\x12\x39\n\texit_type\x18\x13 \x01(\x0e\x32\x17.ray.rpc.WorkerExitTypeH\x00R\x08\x65xitType\x88\x01\x01\x12$\n\x0b\x65xit_detail\x18\x14 \x01(\tH\x01R\nexitDetail\x88\x01\x01\x12\x10\n\x03pid\x18\x15 \x01(\rR\x03pid\x12\"\n\rstart_time_ms\x18\x17 \x01(\x04R\x0bstartTimeMs\x12\x1e\n\x0b\x65nd_time_ms\x18\x18 \x01(\x04R\tendTimeMs\x1a=\n\x0fWorkerInfoEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01\x42\x0c\n\n_exit_typeB\x0e\n\x0c_exit_detail\"K\n\x0fWorkerDeltaData\x12\x1b\n\traylet_id\x18\x01 \x01(\x0cR\x08rayletId\x12\x1b\n\tworker_id\x18\x02 \x01(\x0cR\x08workerId\"\x9a\x01\n\x0bResourceMap\x12\x35\n\x05items\x18\x01 \x03(\x0b\x32\x1f.ray.rpc.ResourceMap.ItemsEntryR\x05items\x1aT\n\nItemsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x30\n\x05value\x18\x02 \x01(\x0b\x32\x1a.ray.rpc.ResourceTableDataR\x05value:\x02\x38\x01\"&\n\x0cStoredConfig\x12\x16\n\x06\x63onfig\x18\x01 \x01(\tR\x06\x63onfig\"\xff\x01\n\x12NodeResourceChange\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12^\n\x11updated_resources\x18\x02 \x03(\x0b\x32\x31.ray.rpc.NodeResourceChange.UpdatedResourcesEntryR\x10updatedResources\x12+\n\x11\x64\x65leted_resources\x18\x03 \x03(\tR\x10\x64\x65letedResources\x1a\x43\n\x15UpdatedResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"3\n\rPubSubMessage\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02id\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61ta\"\x90\x01\n\x0eResourceUpdate\x12\x35\n\x06\x63hange\x18\x01 \x01(\x0b\x32\x1b.ray.rpc.NodeResourceChangeH\x00R\x06\x63hange\x12,\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x16.ray.rpc.ResourcesDataH\x00R\x04\x64\x61taB\x19\n\x17resource_change_or_data\"b\n\x1aResourceUsageBroadcastData\x12\x15\n\x06seq_no\x18\x01 \x01(\x03R\x05seqNo\x12-\n\x05\x62\x61tch\x18\x02 \x03(\x0b\x32\x17.ray.rpc.ResourceUpdateR\x05\x62\x61tch\"\xa7\x02\n\x0eResourceDemand\x12\x38\n\x05shape\x18\x01 \x03(\x0b\x32\".ray.rpc.ResourceDemand.ShapeEntryR\x05shape\x12\x39\n\x19num_ready_requests_queued\x18\x02 \x01(\x04R\x16numReadyRequestsQueued\x12\x43\n\x1enum_infeasible_requests_queued\x18\x03 \x01(\x04R\x1bnumInfeasibleRequestsQueued\x12!\n\x0c\x62\x61\x63klog_size\x18\x04 \x01(\x03R\x0b\x62\x61\x63klogSize\x1a\x38\n\nShapeEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"R\n\x0cResourceLoad\x12\x42\n\x10resource_demands\x18\x01 \x03(\x0b\x32\x17.ray.rpc.ResourceDemandR\x0fresourceDemands\"\xc1\t\n\rResourcesData\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12_\n\x13resources_available\x18\x02 \x03(\x0b\x32..ray.rpc.ResourcesData.ResourcesAvailableEntryR\x12resourcesAvailable\x12>\n\x1bresources_available_changed\x18\x03 \x01(\x08R\x19resourcesAvailableChanged\x12S\n\x0fresources_total\x18\x04 \x03(\x0b\x32*.ray.rpc.ResourcesData.ResourcesTotalEntryR\x0eresourcesTotal\x12M\n\rresource_load\x18\x05 \x03(\x0b\x32(.ray.rpc.ResourcesData.ResourceLoadEntryR\x0cresourceLoad\x12\x32\n\x15resource_load_changed\x18\x06 \x01(\x08R\x13resourceLoadChanged\x12J\n\x16resource_load_by_shape\x18\x07 \x01(\x0b\x32\x15.ray.rpc.ResourceLoadR\x13resourceLoadByShape\x12(\n\x10should_global_gc\x18\x08 \x01(\x08R\x0eshouldGlobalGc\x12\x30\n\x14node_manager_address\x18\t \x01(\tR\x12nodeManagerAddress\x12.\n\x13object_pulls_queued\x18\n \x01(\x08R\x11objectPullsQueued\x12\x63\n\x15resources_normal_task\x18\x0b \x03(\x0b\x32/.ray.rpc.ResourcesData.ResourcesNormalTaskEntryR\x13resourcesNormalTask\x12\x41\n\x1dresources_normal_task_changed\x18\x0c \x01(\x08R\x1aresourcesNormalTaskChanged\x12\x45\n\x1fresources_normal_task_timestamp\x18\r \x01(\x03R\x1cresourcesNormalTaskTimestamp\x12\x44\n\x1f\x63luster_full_of_actors_detected\x18\x0e \x01(\x08R\x1b\x63lusterFullOfActorsDetected\x1a\x45\n\x17ResourcesAvailableEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\x41\n\x13ResourcesTotalEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a?\n\x11ResourceLoadEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\x46\n\x18ResourcesNormalTaskEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xe1\x01\n\x16ResourceUsageBatchData\x12,\n\x05\x62\x61tch\x18\x01 \x03(\x0b\x32\x16.ray.rpc.ResourcesDataR\x05\x62\x61tch\x12J\n\x16resource_load_by_shape\x18\x02 \x01(\x0b\x32\x15.ray.rpc.ResourceLoadR\x13resourceLoadByShape\x12M\n\x14placement_group_load\x18\x03 \x01(\x0b\x32\x1b.ray.rpc.PlacementGroupLoadR\x12placementGroupLoad\"h\n\x12PlacementGroupLoad\x12R\n\x14placement_group_data\x18\x01 \x03(\x0b\x32 .ray.rpc.PlacementGroupTableDataR\x12placementGroupData\"\xdc\x04\n\x13PlacementGroupStats\x12?\n\x1c\x63reation_request_received_ns\x18\x02 \x01(\x03R\x19\x63reationRequestReceivedNs\x12;\n\x1ascheduling_started_time_ns\x18\x03 \x01(\x03R\x17schedulingStartedTimeNs\x12\x32\n\x15scheduling_latency_us\x18\x04 \x01(\x03R\x13schedulingLatencyUs\x12\x41\n\x1e\x65nd_to_end_creation_latency_us\x18\x05 \x01(\x03R\x19\x65ndToEndCreationLatencyUs\x12-\n\x12scheduling_attempt\x18\x06 \x01(\rR\x11schedulingAttempt\x12\x33\n\x16highest_retry_delay_ms\x18\x07 \x01(\x01R\x13highestRetryDelayMs\x12W\n\x10scheduling_state\x18\x08 \x01(\x0e\x32,.ray.rpc.PlacementGroupStats.SchedulingStateR\x0fschedulingState\"\x92\x01\n\x0fSchedulingState\x12\n\n\x06QUEUED\x10\x00\x12\x0b\n\x07REMOVED\x10\x01\x12\x16\n\x12SCHEDULING_STARTED\x10\x02\x12\x10\n\x0cNO_RESOURCES\x10\x03\x12\x0e\n\nINFEASIBLE\x10\x04\x12\x1e\n\x1a\x46\x41ILED_TO_COMMIT_RESOURCES\x10\x05\x12\x0c\n\x08\x46INISHED\x10\x06\"\xb6\x05\n\x17PlacementGroupTableData\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12)\n\x07\x62undles\x18\x03 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x07\x62undles\x12\x36\n\x08strategy\x18\x04 \x01(\x0e\x32\x1a.ray.rpc.PlacementStrategyR\x08strategy\x12J\n\x05state\x18\x05 \x01(\x0e\x32\x34.ray.rpc.PlacementGroupTableData.PlacementGroupStateR\x05state\x12$\n\x0e\x63reator_job_id\x18\x06 \x01(\x0cR\x0c\x63reatorJobId\x12(\n\x10\x63reator_actor_id\x18\x07 \x01(\x0cR\x0e\x63reatorActorId\x12(\n\x10\x63reator_job_dead\x18\x08 \x01(\x08R\x0e\x63reatorJobDead\x12,\n\x12\x63reator_actor_dead\x18\t \x01(\x08R\x10\x63reatorActorDead\x12\x1f\n\x0bis_detached\x18\n \x01(\x08R\nisDetached\x12#\n\rray_namespace\x18\x0b \x01(\tR\x0crayNamespace\x12\x32\n\x05stats\x18\x0c \x01(\x0b\x32\x1c.ray.rpc.PlacementGroupStatsR\x05stats\x12\x38\n\x19max_cpu_fraction_per_node\x18\r \x01(\x01R\x15maxCpuFractionPerNode\"N\n\x13PlacementGroupState\x12\x0b\n\x07PENDING\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07REMOVED\x10\x02\x12\x10\n\x0cRESCHEDULING\x10\x03*\xdc\x02\n\x0bTablePrefix\x12\x14\n\x10TABLE_PREFIX_MIN\x10\x00\x12\n\n\x06UNUSED\x10\x01\x12\x08\n\x04TASK\x10\x02\x12\x0f\n\x0bRAYLET_TASK\x10\x03\x12\x08\n\x04NODE\x10\x04\x12\n\n\x06OBJECT\x10\x05\x12\t\n\x05\x41\x43TOR\x10\x06\x12\x0c\n\x08\x46UNCTION\x10\x07\x12\x17\n\x13TASK_RECONSTRUCTION\x10\x08\x12\x18\n\x14RESOURCE_USAGE_BATCH\x10\t\x12\x07\n\x03JOB\x10\n\x12\x0e\n\nTASK_LEASE\x10\x0c\x12\x11\n\rNODE_RESOURCE\x10\r\x12\x10\n\x0c\x44IRECT_ACTOR\x10\x0e\x12\x0b\n\x07WORKERS\x10\x0f\x12\x13\n\x0fINTERNAL_CONFIG\x10\x10\x12\x1c\n\x18PLACEMENT_GROUP_SCHEDULE\x10\x11\x12\x13\n\x0fPLACEMENT_GROUP\x10\x12\x12\x06\n\x02KV\x10\x13\x12\x13\n\x0f\x41\x43TOR_TASK_SPEC\x10\x14*\xbe\x02\n\x0bTablePubsub\x12\x14\n\x10TABLE_PUBSUB_MIN\x10\x00\x12\x0e\n\nNO_PUBLISH\x10\x01\x12\x0f\n\x0bTASK_PUBSUB\x10\x02\x12\x16\n\x12RAYLET_TASK_PUBSUB\x10\x03\x12\x0f\n\x0bNODE_PUBSUB\x10\x04\x12\x11\n\rOBJECT_PUBSUB\x10\x05\x12\x10\n\x0c\x41\x43TOR_PUBSUB\x10\x06\x12\x1f\n\x1bRESOURCE_USAGE_BATCH_PUBSUB\x10\x07\x12\x15\n\x11TASK_LEASE_PUBSUB\x10\x08\x12\x0e\n\nJOB_PUBSUB\x10\t\x12\x18\n\x14NODE_RESOURCE_PUBSUB\x10\n\x12\x17\n\x13\x44IRECT_ACTOR_PUBSUB\x10\x0b\x12\x19\n\x15WORKER_FAILURE_PUBSUB\x10\x0c\x12\x14\n\x10TABLE_PUBSUB_MAX\x10\r*.\n\rGcsChangeMode\x12\x11\n\rAPPEND_OR_ADD\x10\x00\x12\n\n\x06REMOVE\x10\x01\x42\x1d\n\x18io.ray.runtime.generated\xf8\x01\x01\x62\x06proto3')

_TABLEPREFIX = DESCRIPTOR.enum_types_by_name['TablePrefix']
TablePrefix = enum_type_wrapper.EnumTypeWrapper(_TABLEPREFIX)
_TABLEPUBSUB = DESCRIPTOR.enum_types_by_name['TablePubsub']
TablePubsub = enum_type_wrapper.EnumTypeWrapper(_TABLEPUBSUB)
_GCSCHANGEMODE = DESCRIPTOR.enum_types_by_name['GcsChangeMode']
GcsChangeMode = enum_type_wrapper.EnumTypeWrapper(_GCSCHANGEMODE)
TABLE_PREFIX_MIN = 0
UNUSED = 1
TASK = 2
RAYLET_TASK = 3
NODE = 4
OBJECT = 5
ACTOR = 6
FUNCTION = 7
TASK_RECONSTRUCTION = 8
RESOURCE_USAGE_BATCH = 9
JOB = 10
TASK_LEASE = 12
NODE_RESOURCE = 13
DIRECT_ACTOR = 14
WORKERS = 15
INTERNAL_CONFIG = 16
PLACEMENT_GROUP_SCHEDULE = 17
PLACEMENT_GROUP = 18
KV = 19
ACTOR_TASK_SPEC = 20
TABLE_PUBSUB_MIN = 0
NO_PUBLISH = 1
TASK_PUBSUB = 2
RAYLET_TASK_PUBSUB = 3
NODE_PUBSUB = 4
OBJECT_PUBSUB = 5
ACTOR_PUBSUB = 6
RESOURCE_USAGE_BATCH_PUBSUB = 7
TASK_LEASE_PUBSUB = 8
JOB_PUBSUB = 9
NODE_RESOURCE_PUBSUB = 10
DIRECT_ACTOR_PUBSUB = 11
WORKER_FAILURE_PUBSUB = 12
TABLE_PUBSUB_MAX = 13
APPEND_OR_ADD = 0
REMOVE = 1


_GCSENTRY = DESCRIPTOR.message_types_by_name['GcsEntry']
_OBJECTTABLEDATA = DESCRIPTOR.message_types_by_name['ObjectTableData']
_ACTORTABLEDATA = DESCRIPTOR.message_types_by_name['ActorTableData']
_ACTORTABLEDATA_REQUIREDRESOURCESENTRY = _ACTORTABLEDATA.nested_types_by_name['RequiredResourcesEntry']
_ERRORTABLEDATA = DESCRIPTOR.message_types_by_name['ErrorTableData']
_SCHEDULEDATA = DESCRIPTOR.message_types_by_name['ScheduleData']
_SCHEDULEDATA_SCHEDULEPLANENTRY = _SCHEDULEDATA.nested_types_by_name['SchedulePlanEntry']
_PROFILEEVENTENTRY = DESCRIPTOR.message_types_by_name['ProfileEventEntry']
_PROFILEEVENTS = DESCRIPTOR.message_types_by_name['ProfileEvents']
_TASKSTATEUPDATE = DESCRIPTOR.message_types_by_name['TaskStateUpdate']
_TASKEVENTS = DESCRIPTOR.message_types_by_name['TaskEvents']
_TASKEVENTDATA = DESCRIPTOR.message_types_by_name['TaskEventData']
_RESOURCETABLEDATA = DESCRIPTOR.message_types_by_name['ResourceTableData']
_AVAILABLERESOURCES = DESCRIPTOR.message_types_by_name['AvailableResources']
_AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY = _AVAILABLERESOURCES.nested_types_by_name['ResourcesAvailableEntry']
_GCSNODEINFO = DESCRIPTOR.message_types_by_name['GcsNodeInfo']
_GCSNODEINFO_RESOURCESTOTALENTRY = _GCSNODEINFO.nested_types_by_name['ResourcesTotalEntry']
_JOBSAPIINFO = DESCRIPTOR.message_types_by_name['JobsAPIInfo']
_JOBSAPIINFO_METADATAENTRY = _JOBSAPIINFO.nested_types_by_name['MetadataEntry']
_JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY = _JOBSAPIINFO.nested_types_by_name['EntrypointResourcesEntry']
_JOBTABLEDATA = DESCRIPTOR.message_types_by_name['JobTableData']
_WORKERTABLEDATA = DESCRIPTOR.message_types_by_name['WorkerTableData']
_WORKERTABLEDATA_WORKERINFOENTRY = _WORKERTABLEDATA.nested_types_by_name['WorkerInfoEntry']
_WORKERDELTADATA = DESCRIPTOR.message_types_by_name['WorkerDeltaData']
_RESOURCEMAP = DESCRIPTOR.message_types_by_name['ResourceMap']
_RESOURCEMAP_ITEMSENTRY = _RESOURCEMAP.nested_types_by_name['ItemsEntry']
_STOREDCONFIG = DESCRIPTOR.message_types_by_name['StoredConfig']
_NODERESOURCECHANGE = DESCRIPTOR.message_types_by_name['NodeResourceChange']
_NODERESOURCECHANGE_UPDATEDRESOURCESENTRY = _NODERESOURCECHANGE.nested_types_by_name['UpdatedResourcesEntry']
_PUBSUBMESSAGE = DESCRIPTOR.message_types_by_name['PubSubMessage']
_RESOURCEUPDATE = DESCRIPTOR.message_types_by_name['ResourceUpdate']
_RESOURCEUSAGEBROADCASTDATA = DESCRIPTOR.message_types_by_name['ResourceUsageBroadcastData']
_RESOURCEDEMAND = DESCRIPTOR.message_types_by_name['ResourceDemand']
_RESOURCEDEMAND_SHAPEENTRY = _RESOURCEDEMAND.nested_types_by_name['ShapeEntry']
_RESOURCELOAD = DESCRIPTOR.message_types_by_name['ResourceLoad']
_RESOURCESDATA = DESCRIPTOR.message_types_by_name['ResourcesData']
_RESOURCESDATA_RESOURCESAVAILABLEENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesAvailableEntry']
_RESOURCESDATA_RESOURCESTOTALENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesTotalEntry']
_RESOURCESDATA_RESOURCELOADENTRY = _RESOURCESDATA.nested_types_by_name['ResourceLoadEntry']
_RESOURCESDATA_RESOURCESNORMALTASKENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesNormalTaskEntry']
_RESOURCEUSAGEBATCHDATA = DESCRIPTOR.message_types_by_name['ResourceUsageBatchData']
_PLACEMENTGROUPLOAD = DESCRIPTOR.message_types_by_name['PlacementGroupLoad']
_PLACEMENTGROUPSTATS = DESCRIPTOR.message_types_by_name['PlacementGroupStats']
_PLACEMENTGROUPTABLEDATA = DESCRIPTOR.message_types_by_name['PlacementGroupTableData']
_ACTORTABLEDATA_ACTORSTATE = _ACTORTABLEDATA.enum_types_by_name['ActorState']
_GCSNODEINFO_GCSNODESTATE = _GCSNODEINFO.enum_types_by_name['GcsNodeState']
_PLACEMENTGROUPSTATS_SCHEDULINGSTATE = _PLACEMENTGROUPSTATS.enum_types_by_name['SchedulingState']
_PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE = _PLACEMENTGROUPTABLEDATA.enum_types_by_name['PlacementGroupState']
GcsEntry = _reflection.GeneratedProtocolMessageType('GcsEntry', (_message.Message,), {
  'DESCRIPTOR' : _GCSENTRY,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GcsEntry)
  })
_sym_db.RegisterMessage(GcsEntry)

ObjectTableData = _reflection.GeneratedProtocolMessageType('ObjectTableData', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectTableData)
  })
_sym_db.RegisterMessage(ObjectTableData)

ActorTableData = _reflection.GeneratedProtocolMessageType('ActorTableData', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ACTORTABLEDATA_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ActorTableData.RequiredResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _ACTORTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorTableData)
  })
_sym_db.RegisterMessage(ActorTableData)
_sym_db.RegisterMessage(ActorTableData.RequiredResourcesEntry)

ErrorTableData = _reflection.GeneratedProtocolMessageType('ErrorTableData', (_message.Message,), {
  'DESCRIPTOR' : _ERRORTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ErrorTableData)
  })
_sym_db.RegisterMessage(ErrorTableData)

ScheduleData = _reflection.GeneratedProtocolMessageType('ScheduleData', (_message.Message,), {

  'SchedulePlanEntry' : _reflection.GeneratedProtocolMessageType('SchedulePlanEntry', (_message.Message,), {
    'DESCRIPTOR' : _SCHEDULEDATA_SCHEDULEPLANENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ScheduleData.SchedulePlanEntry)
    })
  ,
  'DESCRIPTOR' : _SCHEDULEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ScheduleData)
  })
_sym_db.RegisterMessage(ScheduleData)
_sym_db.RegisterMessage(ScheduleData.SchedulePlanEntry)

ProfileEventEntry = _reflection.GeneratedProtocolMessageType('ProfileEventEntry', (_message.Message,), {
  'DESCRIPTOR' : _PROFILEEVENTENTRY,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ProfileEventEntry)
  })
_sym_db.RegisterMessage(ProfileEventEntry)

ProfileEvents = _reflection.GeneratedProtocolMessageType('ProfileEvents', (_message.Message,), {
  'DESCRIPTOR' : _PROFILEEVENTS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ProfileEvents)
  })
_sym_db.RegisterMessage(ProfileEvents)

TaskStateUpdate = _reflection.GeneratedProtocolMessageType('TaskStateUpdate', (_message.Message,), {
  'DESCRIPTOR' : _TASKSTATEUPDATE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskStateUpdate)
  })
_sym_db.RegisterMessage(TaskStateUpdate)

TaskEvents = _reflection.GeneratedProtocolMessageType('TaskEvents', (_message.Message,), {
  'DESCRIPTOR' : _TASKEVENTS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskEvents)
  })
_sym_db.RegisterMessage(TaskEvents)

TaskEventData = _reflection.GeneratedProtocolMessageType('TaskEventData', (_message.Message,), {
  'DESCRIPTOR' : _TASKEVENTDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskEventData)
  })
_sym_db.RegisterMessage(TaskEventData)

ResourceTableData = _reflection.GeneratedProtocolMessageType('ResourceTableData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCETABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceTableData)
  })
_sym_db.RegisterMessage(ResourceTableData)

AvailableResources = _reflection.GeneratedProtocolMessageType('AvailableResources', (_message.Message,), {

  'ResourcesAvailableEntry' : _reflection.GeneratedProtocolMessageType('ResourcesAvailableEntry', (_message.Message,), {
    'DESCRIPTOR' : _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.AvailableResources.ResourcesAvailableEntry)
    })
  ,
  'DESCRIPTOR' : _AVAILABLERESOURCES,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.AvailableResources)
  })
_sym_db.RegisterMessage(AvailableResources)
_sym_db.RegisterMessage(AvailableResources.ResourcesAvailableEntry)

GcsNodeInfo = _reflection.GeneratedProtocolMessageType('GcsNodeInfo', (_message.Message,), {

  'ResourcesTotalEntry' : _reflection.GeneratedProtocolMessageType('ResourcesTotalEntry', (_message.Message,), {
    'DESCRIPTOR' : _GCSNODEINFO_RESOURCESTOTALENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.GcsNodeInfo.ResourcesTotalEntry)
    })
  ,
  'DESCRIPTOR' : _GCSNODEINFO,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GcsNodeInfo)
  })
_sym_db.RegisterMessage(GcsNodeInfo)
_sym_db.RegisterMessage(GcsNodeInfo.ResourcesTotalEntry)

JobsAPIInfo = _reflection.GeneratedProtocolMessageType('JobsAPIInfo', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBSAPIINFO_METADATAENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo.MetadataEntry)
    })
  ,

  'EntrypointResourcesEntry' : _reflection.GeneratedProtocolMessageType('EntrypointResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo.EntrypointResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _JOBSAPIINFO,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo)
  })
_sym_db.RegisterMessage(JobsAPIInfo)
_sym_db.RegisterMessage(JobsAPIInfo.MetadataEntry)
_sym_db.RegisterMessage(JobsAPIInfo.EntrypointResourcesEntry)

JobTableData = _reflection.GeneratedProtocolMessageType('JobTableData', (_message.Message,), {
  'DESCRIPTOR' : _JOBTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobTableData)
  })
_sym_db.RegisterMessage(JobTableData)

WorkerTableData = _reflection.GeneratedProtocolMessageType('WorkerTableData', (_message.Message,), {

  'WorkerInfoEntry' : _reflection.GeneratedProtocolMessageType('WorkerInfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _WORKERTABLEDATA_WORKERINFOENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.WorkerTableData.WorkerInfoEntry)
    })
  ,
  'DESCRIPTOR' : _WORKERTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerTableData)
  })
_sym_db.RegisterMessage(WorkerTableData)
_sym_db.RegisterMessage(WorkerTableData.WorkerInfoEntry)

WorkerDeltaData = _reflection.GeneratedProtocolMessageType('WorkerDeltaData', (_message.Message,), {
  'DESCRIPTOR' : _WORKERDELTADATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerDeltaData)
  })
_sym_db.RegisterMessage(WorkerDeltaData)

ResourceMap = _reflection.GeneratedProtocolMessageType('ResourceMap', (_message.Message,), {

  'ItemsEntry' : _reflection.GeneratedProtocolMessageType('ItemsEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEMAP_ITEMSENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMap.ItemsEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCEMAP,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMap)
  })
_sym_db.RegisterMessage(ResourceMap)
_sym_db.RegisterMessage(ResourceMap.ItemsEntry)

StoredConfig = _reflection.GeneratedProtocolMessageType('StoredConfig', (_message.Message,), {
  'DESCRIPTOR' : _STOREDCONFIG,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.StoredConfig)
  })
_sym_db.RegisterMessage(StoredConfig)

NodeResourceChange = _reflection.GeneratedProtocolMessageType('NodeResourceChange', (_message.Message,), {

  'UpdatedResourcesEntry' : _reflection.GeneratedProtocolMessageType('UpdatedResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceChange.UpdatedResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _NODERESOURCECHANGE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceChange)
  })
_sym_db.RegisterMessage(NodeResourceChange)
_sym_db.RegisterMessage(NodeResourceChange.UpdatedResourcesEntry)

PubSubMessage = _reflection.GeneratedProtocolMessageType('PubSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubSubMessage)
  })
_sym_db.RegisterMessage(PubSubMessage)

ResourceUpdate = _reflection.GeneratedProtocolMessageType('ResourceUpdate', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUPDATE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUpdate)
  })
_sym_db.RegisterMessage(ResourceUpdate)

ResourceUsageBroadcastData = _reflection.GeneratedProtocolMessageType('ResourceUsageBroadcastData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUSAGEBROADCASTDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUsageBroadcastData)
  })
_sym_db.RegisterMessage(ResourceUsageBroadcastData)

ResourceDemand = _reflection.GeneratedProtocolMessageType('ResourceDemand', (_message.Message,), {

  'ShapeEntry' : _reflection.GeneratedProtocolMessageType('ShapeEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEDEMAND_SHAPEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceDemand.ShapeEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCEDEMAND,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceDemand)
  })
_sym_db.RegisterMessage(ResourceDemand)
_sym_db.RegisterMessage(ResourceDemand.ShapeEntry)

ResourceLoad = _reflection.GeneratedProtocolMessageType('ResourceLoad', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCELOAD,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceLoad)
  })
_sym_db.RegisterMessage(ResourceLoad)

ResourcesData = _reflection.GeneratedProtocolMessageType('ResourcesData', (_message.Message,), {

  'ResourcesAvailableEntry' : _reflection.GeneratedProtocolMessageType('ResourcesAvailableEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESAVAILABLEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesAvailableEntry)
    })
  ,

  'ResourcesTotalEntry' : _reflection.GeneratedProtocolMessageType('ResourcesTotalEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESTOTALENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesTotalEntry)
    })
  ,

  'ResourceLoadEntry' : _reflection.GeneratedProtocolMessageType('ResourceLoadEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCELOADENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourceLoadEntry)
    })
  ,

  'ResourcesNormalTaskEntry' : _reflection.GeneratedProtocolMessageType('ResourcesNormalTaskEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESNORMALTASKENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesNormalTaskEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCESDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData)
  })
_sym_db.RegisterMessage(ResourcesData)
_sym_db.RegisterMessage(ResourcesData.ResourcesAvailableEntry)
_sym_db.RegisterMessage(ResourcesData.ResourcesTotalEntry)
_sym_db.RegisterMessage(ResourcesData.ResourceLoadEntry)
_sym_db.RegisterMessage(ResourcesData.ResourcesNormalTaskEntry)

ResourceUsageBatchData = _reflection.GeneratedProtocolMessageType('ResourceUsageBatchData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUSAGEBATCHDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUsageBatchData)
  })
_sym_db.RegisterMessage(ResourceUsageBatchData)

PlacementGroupLoad = _reflection.GeneratedProtocolMessageType('PlacementGroupLoad', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPLOAD,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupLoad)
  })
_sym_db.RegisterMessage(PlacementGroupLoad)

PlacementGroupStats = _reflection.GeneratedProtocolMessageType('PlacementGroupStats', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSTATS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupStats)
  })
_sym_db.RegisterMessage(PlacementGroupStats)

PlacementGroupTableData = _reflection.GeneratedProtocolMessageType('PlacementGroupTableData', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupTableData)
  })
_sym_db.RegisterMessage(PlacementGroupTableData)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030io.ray.runtime.generated\370\001\001'
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._options = None
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _SCHEDULEDATA_SCHEDULEPLANENTRY._options = None
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_options = b'8\001'
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._options = None
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_options = b'8\001'
  _GCSNODEINFO_RESOURCESTOTALENTRY._options = None
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_options = b'8\001'
  _JOBSAPIINFO_METADATAENTRY._options = None
  _JOBSAPIINFO_METADATAENTRY._serialized_options = b'8\001'
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._options = None
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_options = b'8\001'
  _WORKERTABLEDATA_WORKERINFOENTRY._options = None
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_options = b'8\001'
  _RESOURCEMAP_ITEMSENTRY._options = None
  _RESOURCEMAP_ITEMSENTRY._serialized_options = b'8\001'
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._options = None
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_options = b'8\001'
  _RESOURCEDEMAND_SHAPEENTRY._options = None
  _RESOURCEDEMAND_SHAPEENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._options = None
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESTOTALENTRY._options = None
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCELOADENTRY._options = None
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._options = None
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_options = b'8\001'
  _TABLEPREFIX._serialized_start=10221
  _TABLEPREFIX._serialized_end=10569
  _TABLEPUBSUB._serialized_start=10572
  _TABLEPUBSUB._serialized_end=10890
  _GCSCHANGEMODE._serialized_start=10892
  _GCSCHANGEMODE._serialized_end=10938
  _GCSENTRY._serialized_start=70
  _GCSENTRY._serialized_end=179
  _OBJECTTABLEDATA._serialized_start=181
  _OBJECTTABLEDATA._serialized_end=224
  _ACTORTABLEDATA._serialized_start=227
  _ACTORTABLEDATA._serialized_end=1485
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_start=1283
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_end=1351
  _ACTORTABLEDATA_ACTORSTATE._serialized_start=1353
  _ACTORTABLEDATA_ACTORSTATE._serialized_end=1450
  _ERRORTABLEDATA._serialized_start=1487
  _ERRORTABLEDATA._serialized_end=1613
  _SCHEDULEDATA._serialized_start=1616
  _SCHEDULEDATA._serialized_end=1773
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_start=1710
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_end=1773
  _PROFILEEVENTENTRY._serialized_start=1776
  _PROFILEEVENTENTRY._serialized_end=1935
  _PROFILEEVENTS._serialized_start=1938
  _PROFILEEVENTS._serialized_end=2119
  _TASKSTATEUPDATE._serialized_start=2122
  _TASKSTATEUPDATE._serialized_end=2646
  _TASKEVENTS._serialized_start=2649
  _TASKEVENTS._serialized_end=2993
  _TASKEVENTDATA._serialized_start=2996
  _TASKEVENTDATA._serialized_end=3208
  _RESOURCETABLEDATA._serialized_start=3210
  _RESOURCETABLEDATA._serialized_end=3274
  _AVAILABLERESOURCES._serialized_start=3277
  _AVAILABLERESOURCES._serialized_end=3495
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_start=3426
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_end=3495
  _GCSNODEINFO._serialized_start=3498
  _GCSNODEINFO._serialized_end=4222
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_start=4120
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_end=4185
  _GCSNODEINFO_GCSNODESTATE._serialized_start=4187
  _GCSNODEINFO_GCSNODESTATE._serialized_end=4222
  _JOBSAPIINFO._serialized_start=4225
  _JOBSAPIINFO._serialized_end=5157
  _JOBSAPIINFO_METADATAENTRY._serialized_start=4853
  _JOBSAPIINFO_METADATAENTRY._serialized_end=4912
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_start=4914
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_end=4984
  _JOBTABLEDATA._serialized_start=5160
  _JOBTABLEDATA._serialized_end=5528
  _WORKERTABLEDATA._serialized_start=5531
  _WORKERTABLEDATA._serialized_end=6146
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_start=6055
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_end=6116
  _WORKERDELTADATA._serialized_start=6148
  _WORKERDELTADATA._serialized_end=6223
  _RESOURCEMAP._serialized_start=6226
  _RESOURCEMAP._serialized_end=6380
  _RESOURCEMAP_ITEMSENTRY._serialized_start=6296
  _RESOURCEMAP_ITEMSENTRY._serialized_end=6380
  _STOREDCONFIG._serialized_start=6382
  _STOREDCONFIG._serialized_end=6420
  _NODERESOURCECHANGE._serialized_start=6423
  _NODERESOURCECHANGE._serialized_end=6678
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_start=6611
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_end=6678
  _PUBSUBMESSAGE._serialized_start=6680
  _PUBSUBMESSAGE._serialized_end=6731
  _RESOURCEUPDATE._serialized_start=6734
  _RESOURCEUPDATE._serialized_end=6878
  _RESOURCEUSAGEBROADCASTDATA._serialized_start=6880
  _RESOURCEUSAGEBROADCASTDATA._serialized_end=6978
  _RESOURCEDEMAND._serialized_start=6981
  _RESOURCEDEMAND._serialized_end=7276
  _RESOURCEDEMAND_SHAPEENTRY._serialized_start=7220
  _RESOURCEDEMAND_SHAPEENTRY._serialized_end=7276
  _RESOURCELOAD._serialized_start=7278
  _RESOURCELOAD._serialized_end=7360
  _RESOURCESDATA._serialized_start=7363
  _RESOURCESDATA._serialized_end=8580
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_start=3426
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_end=3495
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_start=4120
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_end=4185
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_start=8445
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_end=8508
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_start=8510
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_end=8580
  _RESOURCEUSAGEBATCHDATA._serialized_start=8583
  _RESOURCEUSAGEBATCHDATA._serialized_end=8808
  _PLACEMENTGROUPLOAD._serialized_start=8810
  _PLACEMENTGROUPLOAD._serialized_end=8914
  _PLACEMENTGROUPSTATS._serialized_start=8917
  _PLACEMENTGROUPSTATS._serialized_end=9521
  _PLACEMENTGROUPSTATS_SCHEDULINGSTATE._serialized_start=9375
  _PLACEMENTGROUPSTATS_SCHEDULINGSTATE._serialized_end=9521
  _PLACEMENTGROUPTABLEDATA._serialized_start=9524
  _PLACEMENTGROUPTABLEDATA._serialized_end=10218
  _PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE._serialized_start=10140
  _PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE._serialized_end=10218
# @@protoc_insertion_point(module_scope)
