# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import runtime_env_common_pb2 as src_dot_ray_dot_protobuf_dot_runtime__env__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dsrc/ray/protobuf/common.proto\x12\x07ray.rpc\x1a)src/ray/protobuf/runtime_env_common.proto\"M\n\x1eNodeAffinitySchedulingStrategy\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x12\n\x04soft\x18\x02 \x01(\x08R\x04soft\"\xdf\x01\n PlacementGroupSchedulingStrategy\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12?\n\x1cplacement_group_bundle_index\x18\x02 \x01(\x03R\x19placementGroupBundleIndex\x12L\n#placement_group_capture_child_tasks\x18\x03 \x01(\x08R\x1fplacementGroupCaptureChildTasks\"\x1b\n\x19\x44\x65\x66\x61ultSchedulingStrategy\"\x1a\n\x18SpreadSchedulingStrategy\"\xe6\x03\n\x12SchedulingStrategy\x12\x64\n\x1b\x64\x65\x66\x61ult_scheduling_strategy\x18\x01 \x01(\x0b\x32\".ray.rpc.DefaultSchedulingStrategyH\x00R\x19\x64\x65\x66\x61ultSchedulingStrategy\x12z\n#placement_group_scheduling_strategy\x18\x02 \x01(\x0b\x32).ray.rpc.PlacementGroupSchedulingStrategyH\x00R placementGroupSchedulingStrategy\x12\x61\n\x1aspread_scheduling_strategy\x18\x03 \x01(\x0b\x32!.ray.rpc.SpreadSchedulingStrategyH\x00R\x18spreadSchedulingStrategy\x12t\n!node_affinity_scheduling_strategy\x18\x04 \x01(\x0b\x32\'.ray.rpc.NodeAffinitySchedulingStrategyH\x00R\x1enodeAffinitySchedulingStrategyB\x15\n\x13scheduling_strategy\"v\n\x07\x41\x64\x64ress\x12\x1b\n\traylet_id\x18\x01 \x01(\x0cR\x08rayletId\x12\x1d\n\nip_address\x18\x02 \x01(\tR\tipAddress\x12\x12\n\x04port\x18\x03 \x01(\x05R\x04port\x12\x1b\n\tworker_id\x18\x04 \x01(\x0cR\x08workerId\"z\n\x16JavaFunctionDescriptor\x12\x1d\n\nclass_name\x18\x01 \x01(\tR\tclassName\x12#\n\rfunction_name\x18\x02 \x01(\tR\x0c\x66unctionName\x12\x1c\n\tsignature\x18\x03 \x01(\tR\tsignature\"\xa4\x01\n\x18PythonFunctionDescriptor\x12\x1f\n\x0bmodule_name\x18\x01 \x01(\tR\nmoduleName\x12\x1d\n\nclass_name\x18\x02 \x01(\tR\tclassName\x12#\n\rfunction_name\x18\x03 \x01(\tR\x0c\x66unctionName\x12#\n\rfunction_hash\x18\x04 \x01(\tR\x0c\x66unctionHash\"s\n\x15\x43ppFunctionDescriptor\x12#\n\rfunction_name\x18\x01 \x01(\tR\x0c\x66unctionName\x12\x16\n\x06\x63\x61ller\x18\x02 \x01(\tR\x06\x63\x61ller\x12\x1d\n\nclass_name\x18\x03 \x01(\tR\tclassName\"\xc5\x02\n\x12\x46unctionDescriptor\x12[\n\x18java_function_descriptor\x18\x01 \x01(\x0b\x32\x1f.ray.rpc.JavaFunctionDescriptorH\x00R\x16javaFunctionDescriptor\x12\x61\n\x1apython_function_descriptor\x18\x02 \x01(\x0b\x32!.ray.rpc.PythonFunctionDescriptorH\x00R\x18pythonFunctionDescriptor\x12X\n\x17\x63pp_function_descriptor\x18\x03 \x01(\x0b\x32\x1e.ray.rpc.CppFunctionDescriptorH\x00R\x15\x63ppFunctionDescriptorB\x15\n\x13\x66unction_descriptor\"\x9f\x01\n\x10\x43oncurrencyGroup\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\'\n\x0fmax_concurrency\x18\x02 \x01(\x05R\x0emaxConcurrency\x12N\n\x14\x66unction_descriptors\x18\x03 \x03(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x13\x66unctionDescriptors\"\xb3\x02\n\x0cRayErrorInfo\x12\x44\n\x10\x61\x63tor_died_error\x18\x02 \x01(\x0b\x32\x18.ray.rpc.ActorDeathCauseH\x00R\x0e\x61\x63torDiedError\x12\x66\n\x1eruntime_env_setup_failed_error\x18\x03 \x01(\x0b\x32 .ray.rpc.RuntimeEnvFailedContextH\x00R\x1aruntimeEnvSetupFailedError\x12%\n\rerror_message\x18\x04 \x01(\tH\x00R\x0c\x65rrorMessage\x12\x36\n\nerror_type\x18\x0b \x01(\x0e\x32\x12.ray.rpc.ErrorTypeH\x01R\terrorType\x88\x01\x01\x42\x07\n\x05\x65rrorB\r\n\x0b_error_type\"\xb0\x02\n\x17OutOfMemoryErrorContext\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\x12\x1b\n\ttask_name\x18\x02 \x01(\tR\x08taskName\x12&\n\x0fnode_ip_address\x18\x03 \x01(\tR\rnodeIpAddress\x12*\n\x11memory_used_bytes\x18\x04 \x01(\x04R\x0fmemoryUsedBytes\x12,\n\x12memory_total_bytes\x18\x05 \x01(\x04R\x10memoryTotalBytes\x12\x32\n\x15memory_usage_fraction\x18\x06 \x01(\x01R\x13memoryUsageFraction\x12)\n\x10memory_threshold\x18\x07 \x01(\x01R\x0fmemoryThreshold\"W\n\x14NodeDiedErrorContext\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12&\n\x0fnode_ip_address\x18\x02 \x01(\tR\rnodeIpAddress\"\xae\x01\n\x0cRayException\x12-\n\x08language\x18\x01 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12\x31\n\x14serialized_exception\x18\x02 \x01(\x0cR\x13serializedException\x12<\n\x1a\x66ormatted_exception_string\x18\x03 \x01(\tR\x18\x66ormattedExceptionString\"\xd2\x03\n\x0f\x41\x63torDeathCause\x12Z\n\x1d\x63reation_task_failure_context\x18\x01 \x01(\x0b\x32\x15.ray.rpc.RayExceptionH\x00R\x1a\x63reationTaskFailureContext\x12_\n\x1aruntime_env_failed_context\x18\x02 \x01(\x0b\x32 .ray.rpc.RuntimeEnvFailedContextH\x00R\x17runtimeEnvFailedContext\x12Y\n\x18\x61\x63tor_died_error_context\x18\x03 \x01(\x0b\x32\x1e.ray.rpc.ActorDiedErrorContextH\x00R\x15\x61\x63torDiedErrorContext\x12\x64\n\x1b\x61\x63tor_unschedulable_context\x18\x04 \x01(\x0b\x32\".ray.rpc.ActorUnschedulableContextH\x00R\x19\x61\x63torUnschedulableContext\x12\x36\n\x0boom_context\x18\x05 \x01(\x0b\x32\x13.ray.rpc.OomContextH\x00R\noomContextB\t\n\x07\x63ontext\">\n\x17RuntimeEnvFailedContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\"@\n\x19\x41\x63torUnschedulableContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\"\xd3\x02\n\x15\x41\x63torDiedErrorContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\x12\x19\n\x08owner_id\x18\x02 \x01(\x0cR\x07ownerId\x12(\n\x10owner_ip_address\x18\x03 \x01(\tR\x0eownerIpAddress\x12&\n\x0fnode_ip_address\x18\x04 \x01(\tR\rnodeIpAddress\x12\x10\n\x03pid\x18\x05 \x01(\rR\x03pid\x12\x12\n\x04name\x18\x06 \x01(\tR\x04name\x12#\n\rray_namespace\x18\x07 \x01(\tR\x0crayNamespace\x12\x1d\n\nclass_name\x18\x08 \x01(\tR\tclassName\x12\x19\n\x08\x61\x63tor_id\x18\t \x01(\x0cR\x07\x61\x63torId\x12#\n\rnever_started\x18\n \x01(\x08R\x0cneverStarted\"\\\n\nOomContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\x12)\n\x10\x66\x61il_immediately\x18\x02 \x01(\x08R\x0f\x66\x61ilImmediately\"\xef\x03\n\tJobConfig\x12\x1f\n\x0bjvm_options\x18\x02 \x03(\tR\njvmOptions\x12(\n\x10\x63ode_search_path\x18\x03 \x03(\tR\x0e\x63odeSearchPath\x12\x41\n\x10runtime_env_info\x18\x04 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12#\n\rray_namespace\x18\x05 \x01(\tR\x0crayNamespace\x12<\n\x08metadata\x18\x06 \x03(\x0b\x32 .ray.rpc.JobConfig.MetadataEntryR\x08metadata\x12V\n\x16\x64\x65\x66\x61ult_actor_lifetime\x18\x07 \x01(\x0e\x32 .ray.rpc.JobConfig.ActorLifetimeR\x14\x64\x65\x66\x61ultActorLifetime\x12+\n\x12py_driver_sys_path\x18\x08 \x03(\tR\x0fpyDriverSysPath\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"/\n\rActorLifetime\x12\x0c\n\x08\x44\x45TACHED\x10\x00\x12\x10\n\x0cNON_DETACHED\x10\x01\"\x9b\r\n\x08TaskSpec\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x11.ray.rpc.TaskTypeR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12-\n\x08language\x18\x03 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12L\n\x13\x66unction_descriptor\x18\x04 \x01(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x12\x66unctionDescriptor\x12\x15\n\x06job_id\x18\x05 \x01(\x0cR\x05jobId\x12\x17\n\x07task_id\x18\x06 \x01(\x0cR\x06taskId\x12$\n\x0eparent_task_id\x18\x07 \x01(\x0cR\x0cparentTaskId\x12%\n\x0eparent_counter\x18\x08 \x01(\x04R\rparentCounter\x12\x1b\n\tcaller_id\x18\t \x01(\x0cR\x08\x63\x61llerId\x12\x37\n\x0e\x63\x61ller_address\x18\n \x01(\x0b\x32\x10.ray.rpc.AddressR\rcallerAddress\x12$\n\x04\x61rgs\x18\x0b \x03(\x0b\x32\x10.ray.rpc.TaskArgR\x04\x61rgs\x12\x1f\n\x0bnum_returns\x18\x0c \x01(\x04R\nnumReturns\x12W\n\x12required_resources\x18\r \x03(\x0b\x32(.ray.rpc.TaskSpec.RequiredResourcesEntryR\x11requiredResources\x12s\n\x1crequired_placement_resources\x18\x0e \x03(\x0b\x32\x31.ray.rpc.TaskSpec.RequiredPlacementResourcesEntryR\x1arequiredPlacementResources\x12W\n\x18\x61\x63tor_creation_task_spec\x18\x0f \x01(\x0b\x32\x1e.ray.rpc.ActorCreationTaskSpecR\x15\x61\x63torCreationTaskSpec\x12>\n\x0f\x61\x63tor_task_spec\x18\x10 \x01(\x0b\x32\x16.ray.rpc.ActorTaskSpecR\ractorTaskSpec\x12\x1f\n\x0bmax_retries\x18\x11 \x01(\x05R\nmaxRetries\x12%\n\x0eskip_execution\x18\x15 \x01(\x08R\rskipExecution\x12/\n\x13\x64\x65\x62ugger_breakpoint\x18\x16 \x01(\x0cR\x12\x64\x65\x62uggerBreakpoint\x12\x41\n\x10runtime_env_info\x18\x17 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12\x34\n\x16\x63oncurrency_group_name\x18\x18 \x01(\tR\x14\x63oncurrencyGroupName\x12)\n\x10retry_exceptions\x18\x19 \x01(\x08R\x0fretryExceptions\x12O\n$serialized_retry_exception_allowlist\x18\x1a \x01(\x0cR!serializedRetryExceptionAllowlist\x12\x14\n\x05\x64\x65pth\x18\x1b \x01(\x03R\x05\x64\x65pth\x12L\n\x13scheduling_strategy\x18\x1c \x01(\x0b\x32\x1b.ray.rpc.SchedulingStrategyR\x12schedulingStrategy\x12%\n\x0e\x61ttempt_number\x18\x1d \x01(\x04R\rattemptNumber\x12\'\n\x0freturns_dynamic\x18\x1e \x01(\x08R\x0ereturnsDynamic\x12,\n\x12\x64ynamic_return_ids\x18\x1f \x03(\x0cR\x10\x64ynamicReturnIds\x12\x36\n\njob_config\x18  \x01(\x0b\x32\x12.ray.rpc.JobConfigH\x00R\tjobConfig\x88\x01\x01\x12*\n\x11submitter_task_id\x18! \x01(\x0cR\x0fsubmitterTaskId\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1aM\n\x1fRequiredPlacementResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\r\n\x0b_job_config\"\xc4\x05\n\rTaskInfoEntry\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x11.ray.rpc.TaskTypeR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12-\n\x08language\x18\x03 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12+\n\x12\x66unc_or_class_name\x18\x04 \x01(\tR\x0f\x66uncOrClassName\x12>\n\x10scheduling_state\x18\x05 \x01(\x0e\x32\x13.ray.rpc.TaskStatusR\x0fschedulingState\x12\x15\n\x06job_id\x18\x06 \x01(\x0cR\x05jobId\x12\x17\n\x07task_id\x18\x07 \x01(\x0cR\x06taskId\x12$\n\x0eparent_task_id\x18\x08 \x01(\x0cR\x0cparentTaskId\x12\\\n\x12required_resources\x18\r \x03(\x0b\x32-.ray.rpc.TaskInfoEntry.RequiredResourcesEntryR\x11requiredResources\x12\x41\n\x10runtime_env_info\x18\x17 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12\x1c\n\x07node_id\x18\x18 \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x1e\n\x08\x61\x63tor_id\x18\x19 \x01(\x0cH\x01R\x07\x61\x63torId\x88\x01\x01\x12\x31\n\x12placement_group_id\x18\x1a \x01(\x0cH\x02R\x10placementGroupId\x88\x01\x01\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\n\n\x08_node_idB\x0b\n\t_actor_idB\x15\n\x13_placement_group_id\"\xd2\x02\n\x06\x42undle\x12=\n\tbundle_id\x18\x01 \x01(\x0b\x32 .ray.rpc.Bundle.BundleIdentifierR\x08\x62undleId\x12I\n\x0eunit_resources\x18\x02 \x03(\x0b\x32\".ray.rpc.Bundle.UnitResourcesEntryR\runitResources\x12\x17\n\x07node_id\x18\x03 \x01(\x0cR\x06nodeId\x1a\x63\n\x10\x42undleIdentifier\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12!\n\x0c\x62undle_index\x18\x02 \x01(\x05R\x0b\x62undleIndex\x1a@\n\x12UnitResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xbc\x03\n\x12PlacementGroupSpec\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12)\n\x07\x62undles\x18\x03 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x07\x62undles\x12\x36\n\x08strategy\x18\x04 \x01(\x0e\x32\x1a.ray.rpc.PlacementStrategyR\x08strategy\x12$\n\x0e\x63reator_job_id\x18\x05 \x01(\x0cR\x0c\x63reatorJobId\x12(\n\x10\x63reator_actor_id\x18\x06 \x01(\x0cR\x0e\x63reatorActorId\x12(\n\x10\x63reator_job_dead\x18\x07 \x01(\x08R\x0e\x63reatorJobDead\x12,\n\x12\x63reator_actor_dead\x18\x08 \x01(\x08R\x10\x63reatorActorDead\x12\x1f\n\x0bis_detached\x18\t \x01(\x08R\nisDetached\x12\x38\n\x19max_cpu_fraction_per_node\x18\n \x01(\x01R\x15maxCpuFractionPerNode\"\x82\x01\n\x0fObjectReference\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\x12\x35\n\rowner_address\x18\x02 \x01(\x0b\x32\x10.ray.rpc.AddressR\x0cownerAddress\x12\x1b\n\tcall_site\x18\x03 \x01(\tR\x08\x63\x61llSite\"\xbf\x02\n\x14ObjectReferenceCount\x12\x36\n\treference\x18\x01 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\treference\x12\"\n\rhas_local_ref\x18\x02 \x01(\x08R\x0bhasLocalRef\x12.\n\tborrowers\x18\x03 \x03(\x0b\x32\x10.ray.rpc.AddressR\tborrowers\x12\x44\n\x11stored_in_objects\x18\x04 \x03(\x0b\x32\x18.ray.rpc.ObjectReferenceR\x0fstoredInObjects\x12\x39\n\x19\x63ontained_in_borrowed_ids\x18\x05 \x03(\x0cR\x16\x63ontainedInBorrowedIds\x12\x1a\n\x08\x63ontains\x18\x06 \x03(\x0cR\x08\x63ontains\"\xbc\x01\n\x07TaskArg\x12\x37\n\nobject_ref\x18\x01 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\tobjectRef\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61ta\x12\x1a\n\x08metadata\x18\x03 \x01(\x0cR\x08metadata\x12H\n\x13nested_inlined_refs\x18\x04 \x03(\x0b\x32\x18.ray.rpc.ObjectReferenceR\x11nestedInlinedRefs\"\xe8\x04\n\x15\x41\x63torCreationTaskSpec\x12\x19\n\x08\x61\x63tor_id\x18\x02 \x01(\x0cR\x07\x61\x63torId\x12,\n\x12max_actor_restarts\x18\x03 \x01(\x03R\x10maxActorRestarts\x12(\n\x10max_task_retries\x18\x04 \x01(\x03R\x0emaxTaskRetries\x12\x34\n\x16\x64ynamic_worker_options\x18\x05 \x03(\tR\x14\x64ynamicWorkerOptions\x12\'\n\x0fmax_concurrency\x18\x06 \x01(\x05R\x0emaxConcurrency\x12\x1f\n\x0bis_detached\x18\x07 \x01(\x08R\nisDetached\x12\x12\n\x04name\x18\x08 \x01(\tR\x04name\x12#\n\rray_namespace\x18\t \x01(\tR\x0crayNamespace\x12\x1d\n\nis_asyncio\x18\n \x01(\x08R\tisAsyncio\x12%\n\x0e\x65xtension_data\x18\x0b \x01(\tR\rextensionData\x12\x36\n\x17serialized_actor_handle\x18\x0c \x01(\x0cR\x15serializedActorHandle\x12H\n\x12\x63oncurrency_groups\x18\r \x03(\x0b\x32\x19.ray.rpc.ConcurrencyGroupR\x11\x63oncurrencyGroups\x12/\n\x14\x65xecute_out_of_order\x18\x0e \x01(\x08R\x11\x65xecuteOutOfOrder\x12*\n\x11max_pending_calls\x18\x0f \x01(\x05R\x0fmaxPendingCalls\"\x93\x01\n\rActorTaskSpec\x12\x19\n\x08\x61\x63tor_id\x18\x02 \x01(\x0cR\x07\x61\x63torId\x12\x42\n\x1e\x61\x63tor_creation_dummy_object_id\x18\x04 \x01(\x0cR\x1a\x61\x63torCreationDummyObjectId\x12#\n\ractor_counter\x18\x05 \x01(\x04R\x0c\x61\x63torCounter\"6\n\x04Task\x12.\n\ttask_spec\x18\x01 \x01(\x0b\x32\x11.ray.rpc.TaskSpecR\x08taskSpec\">\n\nResourceId\x12\x14\n\x05index\x18\x01 \x01(\x03R\x05index\x12\x1a\n\x08quantity\x18\x02 \x01(\x01R\x08quantity\"^\n\x10ResourceMapEntry\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x0cresource_ids\x18\x02 \x03(\x0b\x32\x13.ray.rpc.ResourceIdR\x0bresourceIds\"\xf4\x03\n\x08ViewData\x12\x1b\n\tview_name\x18\x01 \x01(\tR\x08viewName\x12\x35\n\x08measures\x18\x02 \x03(\x0b\x32\x19.ray.rpc.ViewData.MeasureR\x08measures\x1a\x93\x03\n\x07Measure\x12\x12\n\x04tags\x18\x01 \x01(\tR\x04tags\x12\x1b\n\tint_value\x18\x02 \x01(\x03R\x08intValue\x12!\n\x0c\x64ouble_value\x18\x03 \x01(\x01R\x0b\x64oubleValue\x12)\n\x10\x64istribution_min\x18\x04 \x01(\x01R\x0f\x64istributionMin\x12+\n\x11\x64istribution_mean\x18\x05 \x01(\x01R\x10\x64istributionMean\x12)\n\x10\x64istribution_max\x18\x06 \x01(\x01R\x0f\x64istributionMax\x12-\n\x12\x64istribution_count\x18\x07 \x01(\x01R\x11\x64istributionCount\x12\x44\n\x1e\x64istribution_bucket_boundaries\x18\x08 \x03(\x01R\x1c\x64istributionBucketBoundaries\x12<\n\x1a\x64istribution_bucket_counts\x18\t \x03(\x01R\x18\x64istributionBucketCounts\"\x80\x03\n\rObjectRefInfo\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\x12\x1b\n\tcall_site\x18\x02 \x01(\tR\x08\x63\x61llSite\x12\x1f\n\x0bobject_size\x18\x03 \x01(\x03R\nobjectSize\x12&\n\x0flocal_ref_count\x18\x04 \x01(\x03R\rlocalRefCount\x12\x37\n\x18submitted_task_ref_count\x18\x05 \x01(\x03R\x15submittedTaskRefCount\x12,\n\x12\x63ontained_in_owned\x18\x06 \x03(\x0cR\x10\x63ontainedInOwned\x12(\n\x10pinned_in_memory\x18\x07 \x01(\x08R\x0epinnedInMemory\x12\x34\n\x0btask_status\x18\x08 \x01(\x0e\x32\x13.ray.rpc.TaskStatusR\ntaskStatus\x12%\n\x0e\x61ttempt_number\x18\t \x01(\x04R\rattemptNumber\"\xab\x01\n\x13ResourceAllocations\x12P\n\x0eresource_slots\x18\x01 \x03(\x0b\x32).ray.rpc.ResourceAllocations.ResourceSlotR\rresourceSlots\x1a\x42\n\x0cResourceSlot\x12\x12\n\x04slot\x18\x01 \x01(\x03R\x04slot\x12\x1e\n\nallocation\x18\x02 \x01(\x01R\nallocation\"\xc2\x08\n\x0f\x43oreWorkerStats\x12*\n\x11\x63urrent_task_desc\x18\x01 \x01(\tR\x0f\x63urrentTaskDesc\x12*\n\x11num_pending_tasks\x18\x02 \x01(\x05R\x0fnumPendingTasks\x12\x36\n\x18num_object_refs_in_scope\x18\x03 \x01(\x05R\x14numObjectRefsInScope\x12\x1d\n\nip_address\x18\x07 \x01(\tR\tipAddress\x12\x12\n\x04port\x18\x08 \x01(\x03R\x04port\x12\x19\n\x08\x61\x63tor_id\x18\t \x01(\x0cR\x07\x61\x63torId\x12R\n\x0eused_resources\x18\n \x03(\x0b\x32+.ray.rpc.CoreWorkerStats.UsedResourcesEntryR\rusedResources\x12O\n\rwebui_display\x18\x0b \x03(\x0b\x32*.ray.rpc.CoreWorkerStats.WebuiDisplayEntryR\x0cwebuiDisplay\x12\"\n\rnum_in_plasma\x18\x0c \x01(\x05R\x0bnumInPlasma\x12*\n\x11num_local_objects\x18\r \x01(\x05R\x0fnumLocalObjects\x12\x37\n\x18used_object_store_memory\x18\x0e \x01(\x03R\x15usedObjectStoreMemory\x12*\n\x11task_queue_length\x18\x0f \x01(\x05R\x0ftaskQueueLength\x12,\n\x12num_executed_tasks\x18\x10 \x01(\x05R\x10numExecutedTasks\x12\x1f\n\x0b\x61\x63tor_title\x18\x11 \x01(\tR\nactorTitle\x12\x37\n\x0bobject_refs\x18\x12 \x03(\x0b\x32\x16.ray.rpc.ObjectRefInfoR\nobjectRefs\x12\x15\n\x06job_id\x18\x13 \x01(\x0cR\x05jobId\x12\x1b\n\tworker_id\x18\x14 \x01(\x0cR\x08workerId\x12-\n\x08language\x18\x15 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12\x10\n\x03pid\x18\x16 \x01(\rR\x03pid\x12\x34\n\x0bworker_type\x18\x17 \x01(\x0e\x32\x13.ray.rpc.WorkerTypeR\nworkerType\x12#\n\robjects_total\x18\x18 \x01(\x03R\x0cobjectsTotal\x1a^\n\x12UsedResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.ray.rpc.ResourceAllocationsR\x05value:\x02\x38\x01\x1a?\n\x11WebuiDisplayEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\'\n\x11NodeResourceUsage\x12\x12\n\x04json\x18\x01 \x01(\tR\x04json\"\x87\x02\n\x0bMetricPoint\x12\x1f\n\x0bmetric_name\x18\x01 \x01(\tR\nmetricName\x12\x1c\n\ttimestamp\x18\x02 \x01(\x03R\ttimestamp\x12\x14\n\x05value\x18\x03 \x01(\x01R\x05value\x12\x32\n\x04tags\x18\x04 \x03(\x0b\x32\x1e.ray.rpc.MetricPoint.TagsEntryR\x04tags\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12\x14\n\x05units\x18\x06 \x01(\tR\x05units\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"I\n\x0eNamedActorInfo\x12#\n\rray_namespace\x18\x01 \x01(\tR\x0crayNamespace\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name*)\n\x08Language\x12\n\n\x06PYTHON\x10\x00\x12\x08\n\x04JAVA\x10\x01\x12\x07\n\x03\x43PP\x10\x02*J\n\nWorkerType\x12\n\n\x06WORKER\x10\x00\x12\n\n\x06\x44RIVER\x10\x01\x12\x10\n\x0cSPILL_WORKER\x10\x02\x12\x12\n\x0eRESTORE_WORKER\x10\x03*U\n\x08TaskType\x12\x0f\n\x0bNORMAL_TASK\x10\x00\x12\x17\n\x13\x41\x43TOR_CREATION_TASK\x10\x01\x12\x0e\n\nACTOR_TASK\x10\x02\x12\x0f\n\x0b\x44RIVER_TASK\x10\x03*\x84\x05\n\tErrorType\x12\x0f\n\x0bWORKER_DIED\x10\x00\x12\x0e\n\nACTOR_DIED\x10\x01\x12\x1c\n\x18OBJECT_UNRECONSTRUCTABLE\x10\x02\x12\x1c\n\x18TASK_EXECUTION_EXCEPTION\x10\x03\x12\x14\n\x10OBJECT_IN_PLASMA\x10\x04\x12\x12\n\x0eTASK_CANCELLED\x10\x05\x12\x19\n\x15\x41\x43TOR_CREATION_FAILED\x10\x06\x12\x1c\n\x18RUNTIME_ENV_SETUP_FAILED\x10\x07\x12\x0f\n\x0bOBJECT_LOST\x10\x08\x12\x0e\n\nOWNER_DIED\x10\t\x12\x12\n\x0eOBJECT_DELETED\x10\n\x12 \n\x1c\x44\x45PENDENCY_RESOLUTION_FAILED\x10\x0b\x12\x32\n.OBJECT_UNRECONSTRUCTABLE_MAX_ATTEMPTS_EXCEEDED\x10\x0c\x12,\n(OBJECT_UNRECONSTRUCTABLE_LINEAGE_EVICTED\x10\r\x12\x1a\n\x16OBJECT_FETCH_TIMED_OUT\x10\x0e\x12\x15\n\x11LOCAL_RAYLET_DIED\x10\x0f\x12 \n\x1cTASK_PLACEMENT_GROUP_REMOVED\x10\x10\x12!\n\x1d\x41\x43TOR_PLACEMENT_GROUP_REMOVED\x10\x11\x12\x1c\n\x18TASK_UNSCHEDULABLE_ERROR\x10\x12\x12\x1d\n\x19\x41\x43TOR_UNSCHEDULABLE_ERROR\x10\x13\x12\x15\n\x11OUT_OF_DISK_ERROR\x10\x14\x12\x10\n\x0cOBJECT_FREED\x10\x15\x12\x11\n\rOUT_OF_MEMORY\x10\x16\x12\r\n\tNODE_DIED\x10\x17*\xf4\x01\n\nTaskStatus\x12\x07\n\x03NIL\x10\x00\x12\x16\n\x12PENDING_ARGS_AVAIL\x10\x01\x12\x1b\n\x17PENDING_NODE_ASSIGNMENT\x10\x02\x12\x1f\n\x1bPENDING_OBJ_STORE_MEM_AVAIL\x10\x03\x12\x16\n\x12PENDING_ARGS_FETCH\x10\x04\x12\x17\n\x13SUBMITTED_TO_WORKER\x10\x05\x12\x0b\n\x07RUNNING\x10\x06\x12\x16\n\x12RUNNING_IN_RAY_GET\x10\x07\x12\x17\n\x13RUNNING_IN_RAY_WAIT\x10\x08\x12\x0c\n\x08\x46INISHED\x10\t\x12\n\n\x06\x46\x41ILED\x10\n*|\n\x0eWorkerExitType\x12\x10\n\x0cSYSTEM_ERROR\x10\x00\x12\x18\n\x14INTENDED_SYSTEM_EXIT\x10\x01\x12\x0e\n\nUSER_ERROR\x10\x02\x12\x16\n\x12INTENDED_USER_EXIT\x10\x03\x12\x16\n\x12NODE_OUT_OF_MEMORY\x10\x04*M\n\x11PlacementStrategy\x12\x08\n\x04PACK\x10\x00\x12\n\n\x06SPREAD\x10\x01\x12\x0f\n\x0bSTRICT_PACK\x10\x02\x12\x11\n\rSTRICT_SPREAD\x10\x03\x42\x1d\n\x18io.ray.runtime.generated\xf8\x01\x01\x62\x06proto3')

_LANGUAGE = DESCRIPTOR.enum_types_by_name['Language']
Language = enum_type_wrapper.EnumTypeWrapper(_LANGUAGE)
_WORKERTYPE = DESCRIPTOR.enum_types_by_name['WorkerType']
WorkerType = enum_type_wrapper.EnumTypeWrapper(_WORKERTYPE)
_TASKTYPE = DESCRIPTOR.enum_types_by_name['TaskType']
TaskType = enum_type_wrapper.EnumTypeWrapper(_TASKTYPE)
_ERRORTYPE = DESCRIPTOR.enum_types_by_name['ErrorType']
ErrorType = enum_type_wrapper.EnumTypeWrapper(_ERRORTYPE)
_TASKSTATUS = DESCRIPTOR.enum_types_by_name['TaskStatus']
TaskStatus = enum_type_wrapper.EnumTypeWrapper(_TASKSTATUS)
_WORKEREXITTYPE = DESCRIPTOR.enum_types_by_name['WorkerExitType']
WorkerExitType = enum_type_wrapper.EnumTypeWrapper(_WORKEREXITTYPE)
_PLACEMENTSTRATEGY = DESCRIPTOR.enum_types_by_name['PlacementStrategy']
PlacementStrategy = enum_type_wrapper.EnumTypeWrapper(_PLACEMENTSTRATEGY)
PYTHON = 0
JAVA = 1
CPP = 2
WORKER = 0
DRIVER = 1
SPILL_WORKER = 2
RESTORE_WORKER = 3
NORMAL_TASK = 0
ACTOR_CREATION_TASK = 1
ACTOR_TASK = 2
DRIVER_TASK = 3
WORKER_DIED = 0
ACTOR_DIED = 1
OBJECT_UNRECONSTRUCTABLE = 2
TASK_EXECUTION_EXCEPTION = 3
OBJECT_IN_PLASMA = 4
TASK_CANCELLED = 5
ACTOR_CREATION_FAILED = 6
RUNTIME_ENV_SETUP_FAILED = 7
OBJECT_LOST = 8
OWNER_DIED = 9
OBJECT_DELETED = 10
DEPENDENCY_RESOLUTION_FAILED = 11
OBJECT_UNRECONSTRUCTABLE_MAX_ATTEMPTS_EXCEEDED = 12
OBJECT_UNRECONSTRUCTABLE_LINEAGE_EVICTED = 13
OBJECT_FETCH_TIMED_OUT = 14
LOCAL_RAYLET_DIED = 15
TASK_PLACEMENT_GROUP_REMOVED = 16
ACTOR_PLACEMENT_GROUP_REMOVED = 17
TASK_UNSCHEDULABLE_ERROR = 18
ACTOR_UNSCHEDULABLE_ERROR = 19
OUT_OF_DISK_ERROR = 20
OBJECT_FREED = 21
OUT_OF_MEMORY = 22
NODE_DIED = 23
NIL = 0
PENDING_ARGS_AVAIL = 1
PENDING_NODE_ASSIGNMENT = 2
PENDING_OBJ_STORE_MEM_AVAIL = 3
PENDING_ARGS_FETCH = 4
SUBMITTED_TO_WORKER = 5
RUNNING = 6
RUNNING_IN_RAY_GET = 7
RUNNING_IN_RAY_WAIT = 8
FINISHED = 9
FAILED = 10
SYSTEM_ERROR = 0
INTENDED_SYSTEM_EXIT = 1
USER_ERROR = 2
INTENDED_USER_EXIT = 3
NODE_OUT_OF_MEMORY = 4
PACK = 0
SPREAD = 1
STRICT_PACK = 2
STRICT_SPREAD = 3


_NODEAFFINITYSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['NodeAffinitySchedulingStrategy']
_PLACEMENTGROUPSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['PlacementGroupSchedulingStrategy']
_DEFAULTSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['DefaultSchedulingStrategy']
_SPREADSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['SpreadSchedulingStrategy']
_SCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['SchedulingStrategy']
_ADDRESS = DESCRIPTOR.message_types_by_name['Address']
_JAVAFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['JavaFunctionDescriptor']
_PYTHONFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['PythonFunctionDescriptor']
_CPPFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['CppFunctionDescriptor']
_FUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['FunctionDescriptor']
_CONCURRENCYGROUP = DESCRIPTOR.message_types_by_name['ConcurrencyGroup']
_RAYERRORINFO = DESCRIPTOR.message_types_by_name['RayErrorInfo']
_OUTOFMEMORYERRORCONTEXT = DESCRIPTOR.message_types_by_name['OutOfMemoryErrorContext']
_NODEDIEDERRORCONTEXT = DESCRIPTOR.message_types_by_name['NodeDiedErrorContext']
_RAYEXCEPTION = DESCRIPTOR.message_types_by_name['RayException']
_ACTORDEATHCAUSE = DESCRIPTOR.message_types_by_name['ActorDeathCause']
_RUNTIMEENVFAILEDCONTEXT = DESCRIPTOR.message_types_by_name['RuntimeEnvFailedContext']
_ACTORUNSCHEDULABLECONTEXT = DESCRIPTOR.message_types_by_name['ActorUnschedulableContext']
_ACTORDIEDERRORCONTEXT = DESCRIPTOR.message_types_by_name['ActorDiedErrorContext']
_OOMCONTEXT = DESCRIPTOR.message_types_by_name['OomContext']
_JOBCONFIG = DESCRIPTOR.message_types_by_name['JobConfig']
_JOBCONFIG_METADATAENTRY = _JOBCONFIG.nested_types_by_name['MetadataEntry']
_TASKSPEC = DESCRIPTOR.message_types_by_name['TaskSpec']
_TASKSPEC_REQUIREDRESOURCESENTRY = _TASKSPEC.nested_types_by_name['RequiredResourcesEntry']
_TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY = _TASKSPEC.nested_types_by_name['RequiredPlacementResourcesEntry']
_TASKINFOENTRY = DESCRIPTOR.message_types_by_name['TaskInfoEntry']
_TASKINFOENTRY_REQUIREDRESOURCESENTRY = _TASKINFOENTRY.nested_types_by_name['RequiredResourcesEntry']
_BUNDLE = DESCRIPTOR.message_types_by_name['Bundle']
_BUNDLE_BUNDLEIDENTIFIER = _BUNDLE.nested_types_by_name['BundleIdentifier']
_BUNDLE_UNITRESOURCESENTRY = _BUNDLE.nested_types_by_name['UnitResourcesEntry']
_PLACEMENTGROUPSPEC = DESCRIPTOR.message_types_by_name['PlacementGroupSpec']
_OBJECTREFERENCE = DESCRIPTOR.message_types_by_name['ObjectReference']
_OBJECTREFERENCECOUNT = DESCRIPTOR.message_types_by_name['ObjectReferenceCount']
_TASKARG = DESCRIPTOR.message_types_by_name['TaskArg']
_ACTORCREATIONTASKSPEC = DESCRIPTOR.message_types_by_name['ActorCreationTaskSpec']
_ACTORTASKSPEC = DESCRIPTOR.message_types_by_name['ActorTaskSpec']
_TASK = DESCRIPTOR.message_types_by_name['Task']
_RESOURCEID = DESCRIPTOR.message_types_by_name['ResourceId']
_RESOURCEMAPENTRY = DESCRIPTOR.message_types_by_name['ResourceMapEntry']
_VIEWDATA = DESCRIPTOR.message_types_by_name['ViewData']
_VIEWDATA_MEASURE = _VIEWDATA.nested_types_by_name['Measure']
_OBJECTREFINFO = DESCRIPTOR.message_types_by_name['ObjectRefInfo']
_RESOURCEALLOCATIONS = DESCRIPTOR.message_types_by_name['ResourceAllocations']
_RESOURCEALLOCATIONS_RESOURCESLOT = _RESOURCEALLOCATIONS.nested_types_by_name['ResourceSlot']
_COREWORKERSTATS = DESCRIPTOR.message_types_by_name['CoreWorkerStats']
_COREWORKERSTATS_USEDRESOURCESENTRY = _COREWORKERSTATS.nested_types_by_name['UsedResourcesEntry']
_COREWORKERSTATS_WEBUIDISPLAYENTRY = _COREWORKERSTATS.nested_types_by_name['WebuiDisplayEntry']
_NODERESOURCEUSAGE = DESCRIPTOR.message_types_by_name['NodeResourceUsage']
_METRICPOINT = DESCRIPTOR.message_types_by_name['MetricPoint']
_METRICPOINT_TAGSENTRY = _METRICPOINT.nested_types_by_name['TagsEntry']
_NAMEDACTORINFO = DESCRIPTOR.message_types_by_name['NamedActorInfo']
_JOBCONFIG_ACTORLIFETIME = _JOBCONFIG.enum_types_by_name['ActorLifetime']
NodeAffinitySchedulingStrategy = _reflection.GeneratedProtocolMessageType('NodeAffinitySchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _NODEAFFINITYSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeAffinitySchedulingStrategy)
  })
_sym_db.RegisterMessage(NodeAffinitySchedulingStrategy)

PlacementGroupSchedulingStrategy = _reflection.GeneratedProtocolMessageType('PlacementGroupSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupSchedulingStrategy)
  })
_sym_db.RegisterMessage(PlacementGroupSchedulingStrategy)

DefaultSchedulingStrategy = _reflection.GeneratedProtocolMessageType('DefaultSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _DEFAULTSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DefaultSchedulingStrategy)
  })
_sym_db.RegisterMessage(DefaultSchedulingStrategy)

SpreadSchedulingStrategy = _reflection.GeneratedProtocolMessageType('SpreadSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _SPREADSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SpreadSchedulingStrategy)
  })
_sym_db.RegisterMessage(SpreadSchedulingStrategy)

SchedulingStrategy = _reflection.GeneratedProtocolMessageType('SchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _SCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SchedulingStrategy)
  })
_sym_db.RegisterMessage(SchedulingStrategy)

Address = _reflection.GeneratedProtocolMessageType('Address', (_message.Message,), {
  'DESCRIPTOR' : _ADDRESS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Address)
  })
_sym_db.RegisterMessage(Address)

JavaFunctionDescriptor = _reflection.GeneratedProtocolMessageType('JavaFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _JAVAFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JavaFunctionDescriptor)
  })
_sym_db.RegisterMessage(JavaFunctionDescriptor)

PythonFunctionDescriptor = _reflection.GeneratedProtocolMessageType('PythonFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _PYTHONFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PythonFunctionDescriptor)
  })
_sym_db.RegisterMessage(PythonFunctionDescriptor)

CppFunctionDescriptor = _reflection.GeneratedProtocolMessageType('CppFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _CPPFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CppFunctionDescriptor)
  })
_sym_db.RegisterMessage(CppFunctionDescriptor)

FunctionDescriptor = _reflection.GeneratedProtocolMessageType('FunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _FUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FunctionDescriptor)
  })
_sym_db.RegisterMessage(FunctionDescriptor)

ConcurrencyGroup = _reflection.GeneratedProtocolMessageType('ConcurrencyGroup', (_message.Message,), {
  'DESCRIPTOR' : _CONCURRENCYGROUP,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConcurrencyGroup)
  })
_sym_db.RegisterMessage(ConcurrencyGroup)

RayErrorInfo = _reflection.GeneratedProtocolMessageType('RayErrorInfo', (_message.Message,), {
  'DESCRIPTOR' : _RAYERRORINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RayErrorInfo)
  })
_sym_db.RegisterMessage(RayErrorInfo)

OutOfMemoryErrorContext = _reflection.GeneratedProtocolMessageType('OutOfMemoryErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _OUTOFMEMORYERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.OutOfMemoryErrorContext)
  })
_sym_db.RegisterMessage(OutOfMemoryErrorContext)

NodeDiedErrorContext = _reflection.GeneratedProtocolMessageType('NodeDiedErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _NODEDIEDERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeDiedErrorContext)
  })
_sym_db.RegisterMessage(NodeDiedErrorContext)

RayException = _reflection.GeneratedProtocolMessageType('RayException', (_message.Message,), {
  'DESCRIPTOR' : _RAYEXCEPTION,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RayException)
  })
_sym_db.RegisterMessage(RayException)

ActorDeathCause = _reflection.GeneratedProtocolMessageType('ActorDeathCause', (_message.Message,), {
  'DESCRIPTOR' : _ACTORDEATHCAUSE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorDeathCause)
  })
_sym_db.RegisterMessage(ActorDeathCause)

RuntimeEnvFailedContext = _reflection.GeneratedProtocolMessageType('RuntimeEnvFailedContext', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVFAILEDCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvFailedContext)
  })
_sym_db.RegisterMessage(RuntimeEnvFailedContext)

ActorUnschedulableContext = _reflection.GeneratedProtocolMessageType('ActorUnschedulableContext', (_message.Message,), {
  'DESCRIPTOR' : _ACTORUNSCHEDULABLECONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorUnschedulableContext)
  })
_sym_db.RegisterMessage(ActorUnschedulableContext)

ActorDiedErrorContext = _reflection.GeneratedProtocolMessageType('ActorDiedErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _ACTORDIEDERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorDiedErrorContext)
  })
_sym_db.RegisterMessage(ActorDiedErrorContext)

OomContext = _reflection.GeneratedProtocolMessageType('OomContext', (_message.Message,), {
  'DESCRIPTOR' : _OOMCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.OomContext)
  })
_sym_db.RegisterMessage(OomContext)

JobConfig = _reflection.GeneratedProtocolMessageType('JobConfig', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBCONFIG_METADATAENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobConfig.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _JOBCONFIG,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobConfig)
  })
_sym_db.RegisterMessage(JobConfig)
_sym_db.RegisterMessage(JobConfig.MetadataEntry)

TaskSpec = _reflection.GeneratedProtocolMessageType('TaskSpec', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKSPEC_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec.RequiredResourcesEntry)
    })
  ,

  'RequiredPlacementResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredPlacementResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec.RequiredPlacementResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _TASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec)
  })
_sym_db.RegisterMessage(TaskSpec)
_sym_db.RegisterMessage(TaskSpec.RequiredResourcesEntry)
_sym_db.RegisterMessage(TaskSpec.RequiredPlacementResourcesEntry)

TaskInfoEntry = _reflection.GeneratedProtocolMessageType('TaskInfoEntry', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKINFOENTRY_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskInfoEntry.RequiredResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _TASKINFOENTRY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskInfoEntry)
  })
_sym_db.RegisterMessage(TaskInfoEntry)
_sym_db.RegisterMessage(TaskInfoEntry.RequiredResourcesEntry)

Bundle = _reflection.GeneratedProtocolMessageType('Bundle', (_message.Message,), {

  'BundleIdentifier' : _reflection.GeneratedProtocolMessageType('BundleIdentifier', (_message.Message,), {
    'DESCRIPTOR' : _BUNDLE_BUNDLEIDENTIFIER,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.Bundle.BundleIdentifier)
    })
  ,

  'UnitResourcesEntry' : _reflection.GeneratedProtocolMessageType('UnitResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _BUNDLE_UNITRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.Bundle.UnitResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _BUNDLE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Bundle)
  })
_sym_db.RegisterMessage(Bundle)
_sym_db.RegisterMessage(Bundle.BundleIdentifier)
_sym_db.RegisterMessage(Bundle.UnitResourcesEntry)

PlacementGroupSpec = _reflection.GeneratedProtocolMessageType('PlacementGroupSpec', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupSpec)
  })
_sym_db.RegisterMessage(PlacementGroupSpec)

ObjectReference = _reflection.GeneratedProtocolMessageType('ObjectReference', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFERENCE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectReference)
  })
_sym_db.RegisterMessage(ObjectReference)

ObjectReferenceCount = _reflection.GeneratedProtocolMessageType('ObjectReferenceCount', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFERENCECOUNT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectReferenceCount)
  })
_sym_db.RegisterMessage(ObjectReferenceCount)

TaskArg = _reflection.GeneratedProtocolMessageType('TaskArg', (_message.Message,), {
  'DESCRIPTOR' : _TASKARG,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskArg)
  })
_sym_db.RegisterMessage(TaskArg)

ActorCreationTaskSpec = _reflection.GeneratedProtocolMessageType('ActorCreationTaskSpec', (_message.Message,), {
  'DESCRIPTOR' : _ACTORCREATIONTASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorCreationTaskSpec)
  })
_sym_db.RegisterMessage(ActorCreationTaskSpec)

ActorTaskSpec = _reflection.GeneratedProtocolMessageType('ActorTaskSpec', (_message.Message,), {
  'DESCRIPTOR' : _ACTORTASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorTaskSpec)
  })
_sym_db.RegisterMessage(ActorTaskSpec)

Task = _reflection.GeneratedProtocolMessageType('Task', (_message.Message,), {
  'DESCRIPTOR' : _TASK,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Task)
  })
_sym_db.RegisterMessage(Task)

ResourceId = _reflection.GeneratedProtocolMessageType('ResourceId', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEID,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceId)
  })
_sym_db.RegisterMessage(ResourceId)

ResourceMapEntry = _reflection.GeneratedProtocolMessageType('ResourceMapEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEMAPENTRY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMapEntry)
  })
_sym_db.RegisterMessage(ResourceMapEntry)

ViewData = _reflection.GeneratedProtocolMessageType('ViewData', (_message.Message,), {

  'Measure' : _reflection.GeneratedProtocolMessageType('Measure', (_message.Message,), {
    'DESCRIPTOR' : _VIEWDATA_MEASURE,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ViewData.Measure)
    })
  ,
  'DESCRIPTOR' : _VIEWDATA,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ViewData)
  })
_sym_db.RegisterMessage(ViewData)
_sym_db.RegisterMessage(ViewData.Measure)

ObjectRefInfo = _reflection.GeneratedProtocolMessageType('ObjectRefInfo', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectRefInfo)
  })
_sym_db.RegisterMessage(ObjectRefInfo)

ResourceAllocations = _reflection.GeneratedProtocolMessageType('ResourceAllocations', (_message.Message,), {

  'ResourceSlot' : _reflection.GeneratedProtocolMessageType('ResourceSlot', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEALLOCATIONS_RESOURCESLOT,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceAllocations.ResourceSlot)
    })
  ,
  'DESCRIPTOR' : _RESOURCEALLOCATIONS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceAllocations)
  })
_sym_db.RegisterMessage(ResourceAllocations)
_sym_db.RegisterMessage(ResourceAllocations.ResourceSlot)

CoreWorkerStats = _reflection.GeneratedProtocolMessageType('CoreWorkerStats', (_message.Message,), {

  'UsedResourcesEntry' : _reflection.GeneratedProtocolMessageType('UsedResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _COREWORKERSTATS_USEDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats.UsedResourcesEntry)
    })
  ,

  'WebuiDisplayEntry' : _reflection.GeneratedProtocolMessageType('WebuiDisplayEntry', (_message.Message,), {
    'DESCRIPTOR' : _COREWORKERSTATS_WEBUIDISPLAYENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats.WebuiDisplayEntry)
    })
  ,
  'DESCRIPTOR' : _COREWORKERSTATS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats)
  })
_sym_db.RegisterMessage(CoreWorkerStats)
_sym_db.RegisterMessage(CoreWorkerStats.UsedResourcesEntry)
_sym_db.RegisterMessage(CoreWorkerStats.WebuiDisplayEntry)

NodeResourceUsage = _reflection.GeneratedProtocolMessageType('NodeResourceUsage', (_message.Message,), {
  'DESCRIPTOR' : _NODERESOURCEUSAGE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceUsage)
  })
_sym_db.RegisterMessage(NodeResourceUsage)

MetricPoint = _reflection.GeneratedProtocolMessageType('MetricPoint', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _METRICPOINT_TAGSENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.MetricPoint.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _METRICPOINT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.MetricPoint)
  })
_sym_db.RegisterMessage(MetricPoint)
_sym_db.RegisterMessage(MetricPoint.TagsEntry)

NamedActorInfo = _reflection.GeneratedProtocolMessageType('NamedActorInfo', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDACTORINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NamedActorInfo)
  })
_sym_db.RegisterMessage(NamedActorInfo)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030io.ray.runtime.generated\370\001\001'
  _JOBCONFIG_METADATAENTRY._options = None
  _JOBCONFIG_METADATAENTRY._serialized_options = b'8\001'
  _TASKSPEC_REQUIREDRESOURCESENTRY._options = None
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._options = None
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_options = b'8\001'
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._options = None
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _BUNDLE_UNITRESOURCESENTRY._options = None
  _BUNDLE_UNITRESOURCESENTRY._serialized_options = b'8\001'
  _COREWORKERSTATS_USEDRESOURCESENTRY._options = None
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_options = b'8\001'
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._options = None
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_options = b'8\001'
  _METRICPOINT_TAGSENTRY._options = None
  _METRICPOINT_TAGSENTRY._serialized_options = b'8\001'
  _LANGUAGE._serialized_start=11733
  _LANGUAGE._serialized_end=11774
  _WORKERTYPE._serialized_start=11776
  _WORKERTYPE._serialized_end=11850
  _TASKTYPE._serialized_start=11852
  _TASKTYPE._serialized_end=11937
  _ERRORTYPE._serialized_start=11940
  _ERRORTYPE._serialized_end=12584
  _TASKSTATUS._serialized_start=12587
  _TASKSTATUS._serialized_end=12831
  _WORKEREXITTYPE._serialized_start=12833
  _WORKEREXITTYPE._serialized_end=12957
  _PLACEMENTSTRATEGY._serialized_start=12959
  _PLACEMENTSTRATEGY._serialized_end=13036
  _NODEAFFINITYSCHEDULINGSTRATEGY._serialized_start=85
  _NODEAFFINITYSCHEDULINGSTRATEGY._serialized_end=162
  _PLACEMENTGROUPSCHEDULINGSTRATEGY._serialized_start=165
  _PLACEMENTGROUPSCHEDULINGSTRATEGY._serialized_end=388
  _DEFAULTSCHEDULINGSTRATEGY._serialized_start=390
  _DEFAULTSCHEDULINGSTRATEGY._serialized_end=417
  _SPREADSCHEDULINGSTRATEGY._serialized_start=419
  _SPREADSCHEDULINGSTRATEGY._serialized_end=445
  _SCHEDULINGSTRATEGY._serialized_start=448
  _SCHEDULINGSTRATEGY._serialized_end=934
  _ADDRESS._serialized_start=936
  _ADDRESS._serialized_end=1054
  _JAVAFUNCTIONDESCRIPTOR._serialized_start=1056
  _JAVAFUNCTIONDESCRIPTOR._serialized_end=1178
  _PYTHONFUNCTIONDESCRIPTOR._serialized_start=1181
  _PYTHONFUNCTIONDESCRIPTOR._serialized_end=1345
  _CPPFUNCTIONDESCRIPTOR._serialized_start=1347
  _CPPFUNCTIONDESCRIPTOR._serialized_end=1462
  _FUNCTIONDESCRIPTOR._serialized_start=1465
  _FUNCTIONDESCRIPTOR._serialized_end=1790
  _CONCURRENCYGROUP._serialized_start=1793
  _CONCURRENCYGROUP._serialized_end=1952
  _RAYERRORINFO._serialized_start=1955
  _RAYERRORINFO._serialized_end=2262
  _OUTOFMEMORYERRORCONTEXT._serialized_start=2265
  _OUTOFMEMORYERRORCONTEXT._serialized_end=2569
  _NODEDIEDERRORCONTEXT._serialized_start=2571
  _NODEDIEDERRORCONTEXT._serialized_end=2658
  _RAYEXCEPTION._serialized_start=2661
  _RAYEXCEPTION._serialized_end=2835
  _ACTORDEATHCAUSE._serialized_start=2838
  _ACTORDEATHCAUSE._serialized_end=3304
  _RUNTIMEENVFAILEDCONTEXT._serialized_start=3306
  _RUNTIMEENVFAILEDCONTEXT._serialized_end=3368
  _ACTORUNSCHEDULABLECONTEXT._serialized_start=3370
  _ACTORUNSCHEDULABLECONTEXT._serialized_end=3434
  _ACTORDIEDERRORCONTEXT._serialized_start=3437
  _ACTORDIEDERRORCONTEXT._serialized_end=3776
  _OOMCONTEXT._serialized_start=3778
  _OOMCONTEXT._serialized_end=3870
  _JOBCONFIG._serialized_start=3873
  _JOBCONFIG._serialized_end=4368
  _JOBCONFIG_METADATAENTRY._serialized_start=4260
  _JOBCONFIG_METADATAENTRY._serialized_end=4319
  _JOBCONFIG_ACTORLIFETIME._serialized_start=4321
  _JOBCONFIG_ACTORLIFETIME._serialized_end=4368
  _TASKSPEC._serialized_start=4371
  _TASKSPEC._serialized_end=6062
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_start=5900
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_end=5968
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_start=5970
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_end=6047
  _TASKINFOENTRY._serialized_start=6065
  _TASKINFOENTRY._serialized_end=6773
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_start=5900
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_end=5968
  _BUNDLE._serialized_start=6776
  _BUNDLE._serialized_end=7114
  _BUNDLE_BUNDLEIDENTIFIER._serialized_start=6949
  _BUNDLE_BUNDLEIDENTIFIER._serialized_end=7048
  _BUNDLE_UNITRESOURCESENTRY._serialized_start=7050
  _BUNDLE_UNITRESOURCESENTRY._serialized_end=7114
  _PLACEMENTGROUPSPEC._serialized_start=7117
  _PLACEMENTGROUPSPEC._serialized_end=7561
  _OBJECTREFERENCE._serialized_start=7564
  _OBJECTREFERENCE._serialized_end=7694
  _OBJECTREFERENCECOUNT._serialized_start=7697
  _OBJECTREFERENCECOUNT._serialized_end=8016
  _TASKARG._serialized_start=8019
  _TASKARG._serialized_end=8207
  _ACTORCREATIONTASKSPEC._serialized_start=8210
  _ACTORCREATIONTASKSPEC._serialized_end=8826
  _ACTORTASKSPEC._serialized_start=8829
  _ACTORTASKSPEC._serialized_end=8976
  _TASK._serialized_start=8978
  _TASK._serialized_end=9032
  _RESOURCEID._serialized_start=9034
  _RESOURCEID._serialized_end=9096
  _RESOURCEMAPENTRY._serialized_start=9098
  _RESOURCEMAPENTRY._serialized_end=9192
  _VIEWDATA._serialized_start=9195
  _VIEWDATA._serialized_end=9695
  _VIEWDATA_MEASURE._serialized_start=9292
  _VIEWDATA_MEASURE._serialized_end=9695
  _OBJECTREFINFO._serialized_start=9698
  _OBJECTREFINFO._serialized_end=10082
  _RESOURCEALLOCATIONS._serialized_start=10085
  _RESOURCEALLOCATIONS._serialized_end=10256
  _RESOURCEALLOCATIONS_RESOURCESLOT._serialized_start=10190
  _RESOURCEALLOCATIONS_RESOURCESLOT._serialized_end=10256
  _COREWORKERSTATS._serialized_start=10259
  _COREWORKERSTATS._serialized_end=11349
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_start=11190
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_end=11284
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_start=11286
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_end=11349
  _NODERESOURCEUSAGE._serialized_start=11351
  _NODERESOURCEUSAGE._serialized_end=11390
  _METRICPOINT._serialized_start=11393
  _METRICPOINT._serialized_end=11656
  _METRICPOINT_TAGSENTRY._serialized_start=11601
  _METRICPOINT_TAGSENTRY._serialized_end=11656
  _NAMEDACTORINFO._serialized_start=11658
  _NAMEDACTORINFO._serialized_end=11731
# @@protoc_insertion_point(module_scope)
